@@ -1,70 +1,51 @@
-{-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE OverloadedStrings #-}
 
 {-
     BNF Converter: C Pretty Printer printer
     Copyright (C) 2004  Author:  Michael Pellauer
 
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
--}
-
-{-
-   **************************************************************
-    BNF Converter Module
-
     Description   : This module generates the C Pretty Printer.
                     It also generates the "show" method for
                     printing an abstract syntax tree.
 
                     The generated files use the Visitor design pattern.
 
-    Author        : Michael Pellauer (pellauer@cs.chalmers.se)
-
-    License       : GPL (GNU General Public License)
-
+    Author        : Michael Pellauer
     Created       : 10 August, 2003
-
     Modified      : 3 September, 2003
                     * Added resizable buffers
-
-   **************************************************************
 -}
 
 module BNFC.Backend.C.CFtoCPrinter (cf2CPrinter) where
 
-import Prelude'
+import Prelude hiding ((<>))
+
+import Data.Bifunctor ( second )
+import Data.Char      ( toLower )
+import Data.Either    ( lefts )
+import Data.Foldable  ( toList )
+import Data.List      ( nub )
 
 import BNFC.CF
-import BNFC.Utils ((+++))
-import BNFC.Backend.Common (renderListSepByPrecedence)
+import BNFC.PrettyPrint
+import BNFC.Utils     ( (+++), uniqOn, unless, unlessNull )
+
+import BNFC.Backend.Common
 import BNFC.Backend.Common.NamedVariables
 import BNFC.Backend.Common.StrUtils (renderCharOrString)
-import Data.List
-import Data.Char(toLower)
-import Data.Either (lefts)
-import BNFC.PrettyPrint
 
 -- | Produces (.h file, .c file).
 
 cf2CPrinter :: CF -> (String, String)
 cf2CPrinter cf = (mkHFile cf groups, mkCFile cf groups)
  where
-    groups = fixCoercions (ruleGroupsInternals cf)
+    groups = fixCoercions $ filterOutDefs $ ruleGroupsInternals cf
+    filterOutDefs = map $ second $ filter $ not . isDefinedRule . funRule
 
 {- **** Header (.h) File Methods **** -}
 
--- | An extremely large function to make the Header File.
+-- | Make the Header File.
 
 mkHFile :: CF -> [(Cat,[Rule])] -> String
 mkHFile cf groups = unlines
@@ -77,7 +58,7 @@ mkHFile cf groups = unlines
   footer
  ]
  where
-  eps = nub . map normCat $ allEntryPoints cf
+  eps = nub . map normCat . toList $ allEntryPoints cf
   prPrints s | normCat s == s = "char *print" ++ s' ++ "(" ++ s' ++ " p);\n"
     where
       s' = identCat s
@@ -94,7 +75,7 @@ mkHFile cf groups = unlines
     "#include \"Absyn.h\"",
     "",
     "/* Certain applications may improve performance by changing the buffer size */",
-    "#define BUFFER_INITIAL 2000",
+    "#define BUFFER_INITIAL 2048",
     "/* You may wish to change _L_PAREN or _R_PAREN */",
     "#define _L_PAREN '('",
     "#define _R_PAREN ')'",
@@ -105,6 +86,9 @@ mkHFile cf groups = unlines
     "void renderCS(String s);",
     "void indent(void);",
     "void backup(void);",
+    "void onEmptyLine(void);",
+    "void removeTrailingSpaces(void);",
+    "void removeTrailingWhitespace(void);",
     ""
    ]
   footer = unlines $
@@ -165,49 +149,59 @@ mkCFile cf groups = concat
     footer
    ]
   where
-    eps = nub . map normCat $ allEntryPoints cf
+    eps = nub . map normCat . toList $ allEntryPoints cf
     header = unlines
      [
-      "/*** BNFC-Generated Pretty Printer and Abstract Syntax Viewer ***/",
+      "/*** Pretty Printer and Abstract Syntax Viewer ***/",
       "",
-      "#include \"Printer.h\"",
+      "#include <ctype.h>   /* isspace */",
+      "#include <stddef.h>  /* size_t */",
       "#include <stdio.h>",
       "#include <string.h>",
       "#include <stdlib.h>",
+      "#include \"Printer.h\"",
       "",
       "#define INDENT_WIDTH 2",
       "",
       "int _n_;",
       "char *buf_;",
-      "int cur_;",
-      "int buf_size;",
+      "size_t cur_;",
+      "size_t buf_size;",
       ""
      ]
     printBasics = unlines
      [
       "void ppInteger(Integer n, int i)",
       "{",
-      "  char tmp[16];",
+      -- https://stackoverflow.com/questions/10536207/ansi-c-maximum-number-of-characters-printing-a-decimal-int
+      -- A buffer of 20 characters is sufficient to print the decimal representation
+      -- of a 64bit integer.  Might not be needed here, but does not hurt.
+      "  char tmp[20];",
       "  sprintf(tmp, \"%d\", n);",
-      "  bufAppendS(tmp);",
+      "  renderS(tmp);",
       "}",
       "void ppDouble(Double d, int i)",
       "{",
-      "  char tmp[16];",
-      "  sprintf(tmp, \"%g\", d);",
-      "  bufAppendS(tmp);",
+      -- https://stackoverflow.com/questions/1701055/what-is-the-maximum-length-in-chars-needed-to-represent-any-double-value
+      -- Recommended buffer size is 24 for doubles (IEEE-754):
+      -- (*) 17 digits for the decimal representation of the integral part
+      -- (*)  5 digits for the exponent
+      "  char tmp[24];",
+      "  sprintf(tmp, \"%.15g\", d);",
+      "  renderS(tmp);",
       "}",
       "void ppChar(Char c, int i)",
       "{",
-      "  bufAppendC('\\'');",
-      "  bufAppendC(c);",
-      "  bufAppendC('\\'');",
+      "  char tmp[4];",
+      "  sprintf(tmp, \"'%c'\", c);",
+      "  renderS(tmp);",
       "}",
       "void ppString(String s, int i)",
       "{",
       "  bufAppendC('\\\"');",
       "  bufAppendS(s);",
       "  bufAppendC('\\\"');",
+      "  bufAppendC(' ');",
       "}",
       "void ppIdent(String s, int i)",
       "{",
@@ -228,14 +222,14 @@ mkCFile cf groups = concat
      [
       "void shInteger(Integer i)",
       "{",
-      "  char tmp[16];",
+      "  char tmp[20];",
       "  sprintf(tmp, \"%d\", i);",
       "  bufAppendS(tmp);",
       "}",
       "void shDouble(Double d)",
       "{",
-      "  char tmp[16];",
-      "  sprintf(tmp, \"%g\", d);",
+      "  char tmp[24];",
+      "  sprintf(tmp, \"%.15g\", d);",
       "  bufAppendS(tmp);",
       "}",
       "void shChar(Char c)",
@@ -273,9 +267,9 @@ mkCFile cf groups = concat
      [
       "void bufAppendS(const char *s)",
       "{",
-      "  int len = strlen(s);",
-      "  int n;",
-      "  while (cur_ + len > buf_size)",
+      "  size_t len = strlen(s);",
+      "  size_t n;",
+      "  while (cur_ + len >= buf_size)",
       "  {",
       "    buf_size *= 2; /* Double the buffer size */",
       "    resizeBuffer();",
@@ -289,7 +283,7 @@ mkCFile cf groups = concat
       "}",
       "void bufAppendC(const char c)",
       "{",
-      "  if (cur_ == buf_size)",
+      "  if (cur_ + 1 >= buf_size)",
       "  {",
       "    buf_size *= 2; /* Double the buffer size */",
       "    resizeBuffer();",
@@ -321,7 +315,7 @@ mkCFile cf groups = concat
       "  buf_ = temp;",
       "}",
       "char *buf_;",
-      "int cur_, buf_size;",
+      "size_t cur_, buf_size;",
       ""
      ]
 
@@ -348,52 +342,56 @@ prPrintFun _ = ""
 -- Generates methods for the Pretty Printer
 
 prPrintData :: (Cat, [Rule]) -> String
-prPrintData (cat, rules) = unlines $
- if isList cat
- then
- [
-  "void pp" ++ cl ++ "("++ cl +++ vname ++ ", int i)",
-  "{",
-  "  while(" ++ vname +++ "!= 0)",
-  "  {",
-  "    if (" ++ vname ++ "->" ++ vname ++ "_ == 0)",
-  "    {",
-  visitMember,
-  optsep,
-  "      " ++ vname +++ "= 0;",
-  "    }",
-  "    else",
-  "    {",
-  visitMember,
-  render (nest 6 (renderListSepByPrecedence "i" renderX
-      (getSeparatorByPrecedence rules))),
-  "      " ++ vname +++ "=" +++ vname ++ "->" ++ vname ++ "_;",
-  "    }",
-  "  }",
-  "}",
-  ""
- ] --Not a list:
- else
- [
-   "void pp" ++ cl ++ "(" ++ cl ++ " p, int _i_)",
-   "{",
-   "  switch(p->kind)",
-   "  {",
-   concatMap prPrintRule rules,
-   "  default:",
-   "    fprintf(stderr, \"Error: bad kind field when printing " ++ show cat ++ "!\\n\");",
-   "    exit(1);",
-   "  }",
-   "}\n"
- ]
+prPrintData (cat, rules)
+  | isList cat = unlines $ concat
+    [ [ "void pp" ++ cl ++ "("++ cl +++ vname ++ ", int i)"
+      , "{"
+      , "  if (" ++ vname +++ "== 0)"
+      , "  { /* nil */"
+      ]
+    , unlessNull (swRules isNilFun) $ \ docs ->
+      [ render $ nest 4 $ vcat docs ]
+    , [ "  }" ]
+    , unlessNull (swRules isOneFun) $ \ docs ->
+      [ "  else if (" ++ pre ++ vname ++ "_ == 0)"
+      , "  { /* last */"
+      , render $ nest 4 $ vcat docs
+      , "  }"
+      ]
+    , unlessNull (swRules isConsFun) $ \ docs ->
+      [ "  else"
+      , "  { /* cons */"
+      , render $ nest 4 $ vcat docs
+      , "  }"
+      ]
+    , [ "}"
+      , ""
+      ]
+    ]
+  | otherwise = unlines $ concat
+    [ [ "void pp" ++ cl ++ "(" ++ cl ++ " p, int _i_)"
+      , "{"
+      , "  switch(p->kind)"
+      , "  {"
+      ]
+    , concatMap prPrintRule rules
+    , [ "  default:"
+      , "    fprintf(stderr, \"Error: bad kind field when printing " ++ catToStr cat ++ "!\\n\");"
+      , "    exit(1);"
+      , "  }"
+      , "}"
+      , ""
+      ]
+    ]
  where
-   cl = identCat (normCat cat)
-   ecl = identCat (normCatOfList cat)
-   vname = map toLower cl
-   member = map toLower ecl
-   visitMember = "      pp" ++ ecl ++ "(" ++ vname ++ "->" ++ member ++ "_, i);"
-   sep' = getCons rules
-   optsep = if hasOneFunc rules then "" else "      " ++ render (renderX sep') ++ ";"
+   cl          = identCat (normCat cat)
+   vname       = map toLower cl
+   pre         = vname ++ "->"
+   prules      = sortRulesByPrecedence rules
+   swRules f   = switchByPrecedence "i" $
+                   map (second $ sep . map text . prPrintRule_ pre) $
+                     uniqOn fst $ filter f prules
+                     -- Discard duplicates, can only handle one rule per precedence.
 
 -- | Helper function that call the right c function (renderC or renderS) to
 -- render a literal string.
@@ -411,33 +409,37 @@ renderX sep' = "render" <> char sc <> parens (text sep)
 
 -- | Pretty Printer methods for a rule.
 
-prPrintRule :: Rule -> String
-prPrintRule r@(Rule fun _ cats) | not (isCoercion fun) = unlines
-  [
-   "  case is_" ++ fun ++ ":",
-   lparen,
-   cats',
-   rparen,
-   "    break;\n"
+prPrintRule :: Rule -> [String]
+prPrintRule r@(Rule fun _ _ _) = unless (isCoercion fun) $ concat
+  [ [ "  case is_" ++ fnm ++ ":"
+    , "    if (_i_ > " ++ show p ++ ") renderC(_L_PAREN);"
+    ]
+  , map ("    " ++) $ prPrintRule_ pre r
+  , [ "    if (_i_ > " ++ show p ++ ") renderC(_R_PAREN);"
+    , "    break;"
+    , ""
+    ]
   ]
-   where
-    p = precRule r
-    (lparen, rparen) =
-      ("    if (_i_ > " ++ show p ++ ") renderC(_L_PAREN);",
-       "    if (_i_ > " ++ show p ++ ") renderC(_R_PAREN);")
-    cats' = concatMap (prPrintCat fun) (numVars cats)
-prPrintRule _ = ""
+  where
+    p   = precRule r
+    fnm = funName fun
+    pre = concat [ "p->u.", map toLower fnm, "_." ]
+
+-- | Only render the rhs (items) of a rule.
+
+prPrintRule_ :: IsFun a => String -> Rul a -> [String]
+prPrintRule_ pre (Rule _ _ items _) = map (prPrintItem pre) $ numVars items
 
 -- | This goes on to recurse to the instance variables.
 
-prPrintCat :: String -> Either (Cat, Doc) String -> String
-prPrintCat fnm (c) = case c of
-  Right t -> "    " ++ render (renderX t) ++ ";\n"
-  Left (cat, nt) | isTokenCat cat -> "    pp" ++ basicFunName (render nt) ++ "(p->u." ++ v ++ "_." ++ render nt ++ ", " ++ show (precCat cat) ++ ");\n"
-  Left (InternalCat, _) -> "    /* Internal Category */\n"
-  Left (cat, nt) -> "    pp" ++ identCat (normCat cat) ++ "(p->u." ++ v ++ "_." ++ render nt ++ ", " ++ show (precCat cat) ++ ");\n"
- where
-  v = map toLower (normFun fnm)
+prPrintItem :: String -> Either (Cat, Doc) String -> String
+prPrintItem pre = \case
+  Right t -> render (renderX t) ++ ";"
+  Left (cat, nt) -> concat
+    [ "pp"
+    , maybe (identCat $ normCat cat) basicFunName $ maybeTokenCat cat
+    , "(", pre, render nt, ", ", show (precCat cat), ");"
+    ]
 
 {- **** Abstract Syntax Tree Printer **** -}
 
@@ -494,7 +496,7 @@ prShowData (cat, rules) = unlines $
    "  {",
    concatMap prShowRule rules,
    "  default:",
-   "    fprintf(stderr, \"Error: bad kind field when showing " ++ show cat ++ "!\\n\");",
+   "    fprintf(stderr, \"Error: bad kind field when showing " ++ catToStr cat ++ "!\\n\");",
    "    exit(1);",
    "  }",
    "}\n"
@@ -509,23 +511,24 @@ prShowData (cat, rules) = unlines $
 -- | Pretty Printer methods for a rule.
 
 prShowRule :: Rule -> String
-prShowRule (Rule fun _ cats) | not (isCoercion fun) = unlines
+prShowRule (Rule fun _ cats _) | not (isCoercion fun) = unlines
   [
-   "  case is_" ++ fun ++ ":",
+   "  case is_" ++ f ++ ":",
    "  " ++ lparen,
-   "    bufAppendS(\"" ++ fun ++ "\");\n",
+   "    bufAppendS(\"" ++ f ++ "\");\n",
    "  " ++ optspace,
    cats',
    "  " ++ rparen,
    "    break;"
   ]
    where
+    f = funName fun
     (optspace, lparen, rparen) = if allTerms cats
       then ("","","")
       else ("  bufAppendC(' ');\n", "  bufAppendC('(');\n","  bufAppendC(')');\n")
     cats' = if allTerms cats
         then ""
-        else concat (insertSpaces (map (prShowCat fun) (lefts $ numVars cats)))
+        else concat (insertSpaces (map (prShowCat f) (lefts $ numVars cats)))
     insertSpaces [] = []
     insertSpaces (x:[]) = [x]
     insertSpaces (x:xs) = if x == ""
@@ -536,42 +539,37 @@ prShowRule (Rule fun _ cats) | not (isCoercion fun) = unlines
     allTerms (_:zs) = allTerms zs
 prShowRule _ = ""
 
--- | This goes on to recurse to the instance variables.
-
 prShowCat :: Fun -> (Cat, Doc) -> String
-prShowCat fnm c = case c of
-    (cat,nt) | isTokenCat cat ->
-        "    sh" ++ basicFunName (render nt) ++ "(p->u." ++ v ++ "_." ++ render nt ++ ");\n"
-    (InternalCat, _) -> "    /* Internal Category */\n"
-    (cat,nt) ->
-        "    sh" ++ identCat (normCat cat) ++ "(p->u." ++ v ++ "_." ++ render nt ++ ");\n"
-  where v = map toLower (normFun fnm)
+prShowCat fnm (cat, nt) = concat
+  [ "    sh"
+  , maybe (identCat $ normCat cat) basicFunName $ maybeTokenCat cat
+  , "(p->u."
+  , map toLower fnm
+  , "_."
+  , render nt
+  , ");\n"
+  ]
 
 {- **** Helper Functions Section **** -}
 
 -- | The visit-function name of a basic type.
 
-basicFunName :: String -> String
-basicFunName v
-  | "integer_" `isPrefixOf` v = "Integer"
-  | "char_" `isPrefixOf` v    = "Char"
-  | "string_" `isPrefixOf` v  = "String"
-  | "double_" `isPrefixOf` v  = "Double"
-  | "ident_" `isPrefixOf` v   = "Ident"
-  | otherwise = "Ident" --User-defined type
+basicFunName :: TokenCat -> String
+basicFunName k
+  | k `elem` baseTokenCatNames = k
+  | otherwise                  = "Ident"
 
 -- | An extremely simple @renderC@ for terminals.
 
 prRender :: String
-prRender = unlines
-  [
+prRender = unlines $ concat
+  [ [
       "/* You may wish to change the renderC functions */",
       "void renderC(Char c)",
       "{",
       "  if (c == '{')",
       "  {",
-      "     bufAppendC('\\n');",
-      "     indent();",
+      "     onEmptyLine();",
       "     bufAppendC(c);",
       "     _n_ = _n_ + INDENT_WIDTH;",
       "     bufAppendC('\\n');",
@@ -581,64 +579,93 @@ prRender = unlines
       "     bufAppendC(c);",
       "  else if (c == ')' || c == ']')",
       "  {",
-      "     backup();",
+      "     removeTrailingWhitespace();",
       "     bufAppendC(c);",
+      "     bufAppendC(' ');",
       "  }",
       "  else if (c == '}')",
       "  {",
-      "     int t;",
       "     _n_ = _n_ - INDENT_WIDTH;",
-      "     for(t=0; t<INDENT_WIDTH; t++) {",
-      "       backup();",
-      "     }",
+      "     onEmptyLine();",
       "     bufAppendC(c);",
       "     bufAppendC('\\n\');",
       "     indent();",
       "  }",
       "  else if (c == ',')",
       "  {",
-      "     backup();",
+      "     removeTrailingWhitespace();",
       "     bufAppendC(c);",
       "     bufAppendC(' ');",
       "  }",
       "  else if (c == ';')",
       "  {",
-      "     backup();",
+      "     removeTrailingWhitespace();",
       "     bufAppendC(c);",
       "     bufAppendC('\\n');",
       "     indent();",
       "  }",
+      "  else if (c == ' ') bufAppendC(c);",
       "  else if (c == 0) return;",
       "  else",
       "  {",
-      "     bufAppendC(' ');",
       "     bufAppendC(c);",
       "     bufAppendC(' ');",
       "  }",
       "}",
+      "",
+      "int allIsSpace(String s)",
+      "{",
+      "  char c;",
+      "  while ((c = *s++))",
+      "    if (! isspace(c)) return 0;",
+      "  return 1;",
+      "}",
+      "",
       "void renderS(String s)",
       "{",
-      "  if(strlen(s) > 0)",
+      "  if (*s) /* s[0] != '\\0', string s not empty */",
       "  {",
-      "    bufAppendS(s);",
-      "    bufAppendC(' ');",
+      "    if (allIsSpace(s)) {",
+      "      backup();",
+      "      bufAppendS(s);",
+      "    } else {",
+      "      bufAppendS(s);",
+      "      bufAppendC(' ');",
+      "    }",
       "  }",
       "}",
+      "",
       "void indent(void)",
       "{",
       "  int n = _n_;",
-      "  while (n > 0)",
-      "  {",
+      "  while (--n >= 0)",
       "    bufAppendC(' ');",
-      "    n--;",
-      "  }",
       "}",
+      "",
       "void backup(void)",
       "{",
-      "  if (buf_[cur_ - 1] == ' ')",
-      "  {",
-      "    buf_[cur_ - 1] = 0;",
-      "    cur_--;",
-      "  }",
-      "}"
+      "  if (cur_ && buf_[cur_ - 1] == ' ')",
+      "    buf_[--cur_] = 0;",
+      "}",
+      ""
+    ]
+  , [ "void removeTrailingSpaces()"
+    , "{"
+    , "  while (cur_ && buf_[cur_ - 1] == ' ') --cur_;"
+    , "  buf_[cur_] = 0;"
+    , "}"
+    , ""
+    , "void removeTrailingWhitespace()"
+    , "{"
+    , "  while (cur_ && (buf_[cur_ - 1] == ' ' || buf_[cur_ - 1] == '\\n')) --cur_;"
+    , "  buf_[cur_] = 0;"
+    , "}"
+    , ""
+    , "void onEmptyLine()"
+    , "{"
+    , "  removeTrailingSpaces();"
+    , "  if (cur_ && buf_[cur_ - 1 ] != '\\n') bufAppendC('\\n');"
+    , "  indent();"
+    , "}"
+    ]
   ]
