@@ -1,47 +1,20 @@
-{-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE OverloadedStrings #-}
 
 {-
     BNF Converter: C Skeleton generator
     Copyright (C) 2004  Author:  Michael Pellauer
 
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
--}
-
-{-
-   **************************************************************
-    BNF Converter Module
-
     Description   : This module generates the C Skeleton functions.
 
                     The generated files follow Appel's case method.
 
-    Author        : Michael Pellauer (pellauer@cs.chalmers.se)
-
-    License       : GPL (GNU General Public License)
-
+    Author        : Michael Pellauer
     Created       : 9 August, 2003
-
-    Modified      : 12 August, 2003
-
-
-   **************************************************************
 -}
 
 module BNFC.Backend.C.CFtoCSkel (cf2CSkel) where
 
-import Prelude'
+import Prelude hiding ((<>))
 
 import BNFC.CF
 import BNFC.Utils                       ( (+++) )
@@ -55,7 +28,7 @@ import Text.PrettyPrint
 cf2CSkel :: CF -> (String, String)
 cf2CSkel cf = (mkHFile cf groups, mkCFile cf groups)
  where
-    groups = fixCoercions (ruleGroups cf)
+    groups = fixCoercions (ruleGroupsInternals cf)
 
 
 {- **** Header (.H) File Functions **** -}
@@ -113,7 +86,7 @@ mkCFile cf groups = concat
   where
     user = map fst $ tokenPragmas cf
     header = unlines [
-      "/*** BNFC-Generated Visitor Traversal Skeleton. ***/",
+      "/*** Visitor Traversal Skeleton. ***/",
       "/* This traverses the abstract syntax tree.",
       "   To use, copy Skeleton.h and Skeleton.c to",
       "   new files. */",
@@ -192,30 +165,33 @@ prData (cat, rules)
           member = map toLower ecl
 
 -- | Visits all the instance variables of a category.
--- >>> let ab = Cat "ab"
--- >>> prPrintRule (Rule "abc" undefined [Left ab, Left ab])
---   case is_abc:
---     /* Code for abc Goes Here */
---     visitab(p->u.abc_.ab_1);
---     visitab(p->u.abc_.ab_2);
+-- >>> let ab = Cat "Ab"
+-- >>> prPrintRule (Rule "Abc" undefined [Left ab, Left ab] Parsable)
+--   case is_Abc:
+--     /* Code for Abc Goes Here */
+--     visitAb(p->u.abc_.ab_1);
+--     visitAb(p->u.abc_.ab_2);
 --     break;
 -- <BLANKLINE>
--- >>> let ab = TokenCat "ab"
--- >>> prPrintRule (Rule "abc" undefined [Left ab])
---   case is_abc:
---     /* Code for abc Goes Here */
+-- >>> let ab = TokenCat "Ab"
+-- >>> prPrintRule (Rule "Abc" undefined [Left ab] Parsable)
+--   case is_Abc:
+--     /* Code for Abc Goes Here */
 --     visitAb(p->u.abc_.ab_);
 --     break;
 -- <BLANKLINE>
--- >>> prPrintRule (Rule "abc" undefined [Left ab, Left ab])
---   case is_abc:
---     /* Code for abc Goes Here */
+-- >>> prPrintRule (Rule "Abc" undefined [Left ab, Left ab] Parsable)
+--   case is_Abc:
+--     /* Code for Abc Goes Here */
 --     visitAb(p->u.abc_.ab_1);
 --     visitAb(p->u.abc_.ab_2);
 --     break;
 -- <BLANKLINE>
 prPrintRule :: Rule -> Doc
-prPrintRule (Rule fun _c cats) | not (isCoercion fun) = nest 2 $ vcat
+prPrintRule (Rule f _c cats _)
+  | isCoercion f    = empty
+  | isDefinedRule f = empty
+  | otherwise       = nest 2 $ vcat
     [ text $ "case is_" ++ fun ++ ":"
     , nest 2 (vcat
         [ "/* Code for " <> text fun <> " Goes Here */"
@@ -224,8 +200,8 @@ prPrintRule (Rule fun _c cats) | not (isCoercion fun) = nest 2 $ vcat
         ])
     ]
   where
+    fun = funName f
     cats' = vcat $ map (prCat fun) (lefts (numVars cats))
-prPrintRule (Rule _fun _ _) = ""
 
 -- Prints the actual instance-variable visiting.
 prCat :: Fun -> (Cat, Doc) -> Doc
@@ -234,7 +210,7 @@ prCat fnm (cat, vname) =
                        then basicFunName cat
                        else text (identCat (normCat cat))
       in visitf <> parens ("p->u." <> text v <> "_." <> vname ) <> ";"
-    where v = map toLower $ normFun fnm
+    where v = map toLower fnm
 
 -- | The visit-function name of a basic type
 
