@@ -1,4 +1,5 @@
-{-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE OverloadedStrings #-}
 
 {-
    **************************************************************
@@ -10,12 +11,8 @@
 
                     The generated files use the Visitor design pattern.
 
-    Author        : Michael Pellauer (pellauer@cs.chalmers.se)
-
-    License       : GPL (GNU General Public License)
-
+    Author        : Michael Pellauer
     Created       : 10 August, 2003
-
     Modified      : 3 September, 2003
                     * Added resizable buffers
 
@@ -24,14 +21,16 @@
 
 module BNFC.Backend.CPP.PrettyPrinter (cf2CPPPrinter, prRender) where
 
-import Prelude'
+import Prelude hiding ((<>))
+
+import Data.Bifunctor (second)
+import Data.Char  (toLower)
 
 import BNFC.CF
-import BNFC.Utils ((+++))
+import BNFC.Utils
 import BNFC.Backend.Common
 import BNFC.Backend.Common.NamedVariables
 import BNFC.Backend.Common.StrUtils (renderCharOrString)
-import Data.Char(toLower)
 import BNFC.Backend.CPP.STL.STLUtils
 import BNFC.PrettyPrint
 
@@ -40,11 +39,12 @@ cf2CPPPrinter :: Bool -> Maybe String -> CF -> (String, String)
 cf2CPPPrinter useStl inPackage cf =
     (mkHFile useStl inPackage cf groups, mkCFile useStl inPackage cf groups)
  where
-    groups = positionRules cf ++ fixCoercions (ruleGroupsInternals cf)
+    groups = when useStl (positionRules cf)  -- CPP/NoSTL treats position tokens as just tokens
+          ++ fixCoercions (ruleGroupsInternals cf)
 
 positionRules :: CF -> [(Cat,[Rule])]
 positionRules cf =
-  [ (TokenCat cat, [Rule cat (TokenCat cat) $ map (Left . TokenCat) [catString, catInteger]])
+  [ (TokenCat cat, [ Rule (noPosition cat) (noPosition $ TokenCat cat) (map (Left . TokenCat) [catString, catInteger]) Parsable ])
   | cat <- filter (isPositionCat cf) $ map fst (tokenPragmas cf)
   ]
 
@@ -53,16 +53,15 @@ positionRules cf =
 --An extremely large function to make the Header File
 mkHFile :: Bool -> Maybe String -> CF -> [(Cat,[Rule])] -> String
 mkHFile useStl inPackage cf groups = unlines
- [
-  printHeader,
-  concatMap prDataH groups,
-  classFooter,
-  showHeader,
-  concatMap prDataH groups,
-  classFooter,
-  footer
- ]
- where
+  [ printHeader
+  , content
+  , classFooter
+  , showHeader
+  , content
+  , classFooter
+  , footer
+  ]
+  where
   printHeader = unlines
    [
     "#ifndef " ++ hdef,
@@ -70,6 +69,7 @@ mkHFile useStl inPackage cf groups = unlines
     "",
     "#include \"Absyn.H\"",
     "#include <stdio.h>",
+    "#include <stddef.h>",
     "#include <string.h>",
     "#include <stdlib.h>",
     "",
@@ -91,12 +91,16 @@ mkHFile useStl inPackage cf groups = unlines
     "  void render(const char *s);",
     "  void indent(void);",
     "  void backup(void);",
+    "  void onEmptyLine(void);",
+    "  void removeTrailingSpaces(void);",
+    "  void removeTrailingWhitespace(void);",
     " public:",
     "  PrintAbsyn(void);",
     "  ~PrintAbsyn(void);",
     "  char *print(Visitable *v);"
    ]
   hdef = nsDefine inPackage "PRINTER_HEADER"
+  content = concatMap (prDataH useStl) groups
   classFooter = unlines $
    [
     "  void visitInteger(Integer i);",
@@ -108,62 +112,48 @@ mkHFile useStl inPackage cf groups = unlines
    [
     " protected:",
     "  char *buf_;",
-    "  int cur_, buf_size;",
+    "  size_t cur_, buf_size;",
     "",
     "  void inline bufAppend(const char *s)",
     "  {",
-    "    int len = strlen(s);",
-    "    while (cur_ + len >= buf_size)",
-    "    {",
-    "      buf_size *= 2; /* Double the buffer size */",
+    "    size_t end = cur_ + strlen(s);",
+    "    if (end >= buf_size) {",
+    "      do buf_size *= 2; /* Double the buffer size */",
+    "      while (end >= buf_size);",
+    "      resizeBuffer();",
     "    }",
-    "    resizeBuffer();",
-    "    for(int n = 0; n < len; n++)",
-    "    {",
-    "      buf_[cur_ + n] = s[n];",
-    "    }",
-    "    cur_ += len;",
-    "    buf_[cur_] = 0;",
+    "    strcpy(&buf_[cur_], s);",
+    "    cur_ = end;",
     "  }",
     "",
     "  void inline bufAppend(const char c)",
     "  {",
-    "    if (cur_ >= buf_size)",
+    "    if (cur_ + 1 >= buf_size)",
     "    {",
     "      buf_size *= 2; /* Double the buffer size */",
     "      resizeBuffer();",
     "    }",
     "    buf_[cur_] = c;",
-    "    cur_++;",
-    "    buf_[cur_] = 0;",
+    "    buf_[++cur_] = 0;",
     "  }",
     "",
     if useStl then render (nest 2 bufAppendString) else "",
     "  void inline bufReset(void)",
     "  {",
-    "    if (buf_) free(buf_);",
+    "    if (buf_) delete[] buf_;",
     "    buf_size = " ++ nsDefine inPackage "BUFFER_INITIAL" ++ ";",
-    "    buf_ = (char *) malloc(buf_size);",
-    "    if (!buf_) {",
-    "      fprintf(stderr, \"Error: Out of memory while allocating buffer!\\n\");",
-    "      exit(1);",
-    "    }",
+    "    buf_ = new char[buf_size];",
     "    memset(buf_, 0, buf_size);",
     "    cur_ = 0;",
     "  }",
     "",
     "  void inline resizeBuffer(void)",
     "  {",
-    "    char *temp = (char *) malloc(buf_size);",
-    "    if (!temp)",
-    "    {",
-    "      fprintf(stderr, \"Error: Out of memory while attempting to grow buffer!\\n\");",
-    "      exit(1);",
-    "    }",
+    "    char *temp = new char[buf_size];",
     "    if (buf_)",
     "    {",
     "      strcpy(temp, buf_);",
-    "      free(buf_);",
+    "      delete[] buf_;",
     "    }",
     "    buf_ = temp;",
     "  }",
@@ -175,7 +165,8 @@ mkHFile useStl inPackage cf groups = unlines
       "void inline bufAppend(String str)"
       $$ codeblock 2
           [ "const char *s = str.c_str();"
-          , "bufAppend(s);"]
+          , "bufAppend(s);"
+          ]
   showHeader = unlines
    [
     "",
@@ -194,21 +185,25 @@ mkHFile useStl inPackage cf groups = unlines
    ]
 
 --Prints all the required method names and their parameters.
-prDataH :: (Cat, [Rule]) -> String
-prDataH (cat, rules) =
- if isList cat
- then concat ["  void visit", cl, "(", cl, " *p);\n"]
- else abstract ++ concatMap prRuleH rules
+prDataH :: Bool -> (Cat, [Rule]) -> String
+prDataH useSTL (cat, rules)
+ | isList cat = unlines $ concat
+     [ [ concat [ "  void visit", cl, "(", cl, " *p);"          ] ]
+     , when useSTL
+       [ concat [ "  void iter", cl, "(", itty, " i, ", itty, " j);" ] ]
+     ]
+ | otherwise  = abstract ++ concatMap prRuleH rules
  where
-   cl = identCat (normCat cat)
-   abstract = case lookupRule (show cat) rules of
+   cl       = identCat (normCat cat)
+   itty     = concat [ cl, "::", "const_iterator" ]
+   abstract = case lookupRule (noPosition $ catToStr cat) rules of
     Just _ -> ""
     Nothing ->  "  void visit" ++ cl ++ "(" ++ cl ++ " *p); /* abstract class */\n"
 
 --Prints all the methods to visit a rule.
-prRuleH :: Rule -> String
-prRuleH (Rule fun _ _) | isProperLabel fun = concat
-  ["  void visit", fun, "(", fun, " *p);\n"]
+prRuleH :: IsFun f => Rul f -> String
+prRuleH (Rule fun _ _ _) | isProperLabel fun = concat
+  ["  void visit", funName fun, "(", funName fun, " *p);\n"]
 prRuleH _ = ""
 
 {- **** Implementation (.C) File Methods **** -}
@@ -233,7 +228,7 @@ mkCFile useStl inPackage cf groups = concat
   where
     header = unlines
      [
-      "/*** BNFC-Generated Pretty Printer and Abstract Syntax Viewer ***/",
+      "/*** Pretty Printer and Abstract Syntax Viewer ***/",
       "",
       "#include <string>",
       "#include \"Printer.H\"",
@@ -286,23 +281,23 @@ mkCFile useStl inPackage cf groups = concat
      [
       "void PrintAbsyn::visitInteger(Integer i)",
       "{",
-      "  char tmp[16];",
+      "  char tmp[20];",
       "  sprintf(tmp, \"%d\", i);",
-      "  bufAppend(tmp);",
+      "  render(tmp);",
       "}",
       "",
       "void PrintAbsyn::visitDouble(Double d)",
       "{",
-      "  char tmp[16];",
-      "  sprintf(tmp, \"%g\", d);",
-      "  bufAppend(tmp);",
+      "  char tmp[24];",
+      "  sprintf(tmp, \"%.15g\", d);",
+      "  render(tmp);",
       "}",
       "",
       "void PrintAbsyn::visitChar(Char c)",
       "{",
-      "  bufAppend('\\'');",
-      "  bufAppend(c);",
-      "  bufAppend('\\'');",
+      "  char tmp[4];",
+      "  sprintf(tmp, \"'%c'\", c);",
+      "  render(tmp);",
       "}",
       "",
       "void PrintAbsyn::visitString(String s)",
@@ -310,6 +305,7 @@ mkCFile useStl inPackage cf groups = concat
       "  bufAppend('\\\"');",
       "  bufAppend(s);",
       "  bufAppend('\\\"');",
+      "  bufAppend(' ');",
       "}",
       "",
       "void PrintAbsyn::visitIdent(String s)",
@@ -333,14 +329,14 @@ mkCFile useStl inPackage cf groups = concat
      [
       "void ShowAbsyn::visitInteger(Integer i)",
       "{",
-      "  char tmp[16];",
+      "  char tmp[20];",
       "  sprintf(tmp, \"%d\", i);",
       "  bufAppend(tmp);",
       "}",
       "void ShowAbsyn::visitDouble(Double d)",
       "{",
-      "  char tmp[16];",
-      "  sprintf(tmp, \"%g\", d);",
+      "  char tmp[24];",
+      "  sprintf(tmp, \"%.15g\", d);",
       "  bufAppend(tmp);",
       "}",
       "void ShowAbsyn::visitChar(Char c)",
@@ -388,7 +384,7 @@ prPrintData False {- use STL -} _ _ (cat@(ListCat _), rules) =
 -- Not a list :
 prPrintData _ _inPackage cf (TokenCat cat, _rules) | isPositionCat cf cat = unlines $
   -- a position token
-  [ "void PrintAbsyn::visit" ++ show cat ++ "(" ++ show cat ++ " *p)"
+  [ "void PrintAbsyn::visit" ++ cat ++ "(" ++ cat ++ " *p)"
   , "{"
   , "  visitIdent(p->string_);"
   , "}"
@@ -398,135 +394,151 @@ prPrintData _ inPackage _cf (cat, rules) = -- Not a list
     abstract ++ concatMap (prPrintRule inPackage) rules
   where
   cl = identCat (normCat cat)
-  abstract = case lookupRule (show cat) rules of
+  abstract = case lookupRule (noPosition $ catToStr cat) rules of
     Just _ -> ""
     Nothing ->  "void PrintAbsyn::visit" ++ cl ++ "(" ++ cl +++ "*p) {} //abstract class\n\n"
 
--- | Generate pretty printer visitor for a list category:
+-- | Generate pretty printer visitor for a list category (STL version).
 --
--- >>> let c = Cat "C" ; lc = ListCat c
--- >>> let rules = [Rule "[]" lc [], Rule "(:)" lc [Left c, Right "-", Left lc]]
--- >>> genPrintVisitorList (lc, rules)
--- void PrintAbsyn::visitListC(ListC *listc)
--- {
---   for (ListC::const_iterator i = listc->begin() ; i != listc->end() ; ++i)
---   {
---     (*i)->accept(this);
---     render('-');
---   }
--- }
---
--- >>> let c2 = CoercCat "C" 2 ; lc2 = ListCat c2
--- >>> let rules2 = rules ++ [Rule "[]" lc2 [], Rule "(:)" lc2 [Left c2, Right "+", Left lc2]]
--- >>> genPrintVisitorList (lc, rules2)
--- void PrintAbsyn::visitListC(ListC *listc)
--- {
---   for (ListC::const_iterator i = listc->begin() ; i != listc->end() ; ++i)
---   {
---     (*i)->accept(this);
---     switch(_i_)
---     {
---       case 2: render('+'); break;
---       default: render('-');
---     }
---   }
--- }
 genPrintVisitorList :: (Cat, [Rule]) -> Doc
-genPrintVisitorList (cat@(ListCat c), rules) =
-    "void PrintAbsyn::visit" <> text cl <> "("<> text cl <> " *" <> vname <> ")"
-    $$ codeblock 2
-      [ "for ("<> text cl <> "::const_iterator i = " <> vname <> "->begin() ; i != " <> vname <> "->end() ; ++i)"
-      , codeblock 2
-          [ if isTokenCat c
-              then "visit" <> text (baseName cl) <> "(*i) ;"
-              else "(*i)->accept(this);"
-          , (if hasOneFunc rules
-              then "if (i != " <> vname <> "->end() - 1)"
-              else empty)
-            <+> renderListSepByPrecedence "_i_" renderSep separators
-          ]
+genPrintVisitorList (cat@(ListCat _), rules) = vcat
+  [ "void PrintAbsyn::visit" <> lty <> parens (lty <+> "*" <> vname)
+  , codeblock 2
+    [ "iter" <> lty <> parens (vname <> "->begin()" <> comma <+> vname <> "->end()") <> semi ]
+  , ""
+  , "void PrintAbsyn::iter" <> lty <> parens (itty <+> "i" <> comma <+> itty <+> "j")
+  , codeblock 2 $ concat
+    [ if null docs0 then
+      [ "if (i == j) return;" ]
+      else
+      [ "if (i == j)"
+      , "{ /* nil */"
+      , nest 2 $ vcat docs0
+      , "}"
+      , "else"
       ]
- where
-   separators = getSeparatorByPrecedence rules
-   cl = identCat (normCat cat)
-   vname = text $ map toLower cl
-   renderSep s = "render(" <> text (snd (renderCharOrString s)) <> ")"
+    , unless (null docs1)
+      [ "if (i == j-1)"
+      , "{ /* last */"
+      , nest 2 $ vcat docs1
+      , "}"
+      , "else"
+      ]
+    , [ "{ /* cons */"
+      ,  nest 2 $ vcat docs2
+      , "}"
+      ]
+    ]
+  , ""
+  , ""
+  ]
+  where
+  cl        = identCat (normCat cat)
+  lty       = text cl                   -- List type
+  itty      = lty <> "::const_iterator" -- Iterator type
+  vname     = text $ map toLower cl
+  prules    = sortRulesByPrecedence rules
+  swRules f = switchByPrecedence "_i_" $
+                map (second $ sep . prListRule_) $
+                  uniqOn fst $ filter f prules
+                  -- Discard duplicates, can only handle one rule per precedence.
+  docs0     = swRules isNilFun
+  docs1     = swRules isOneFun
+  docs2     = swRules isConsFun
+
 genPrintVisitorList _ = error "genPrintVisitorList expects a ListCat"
 
--- | This is the only part of the pretty printer that differs significantly
+-- | Only render the rhs (items) of a list rule.
+
+prListRule_ :: IsFun a => Rul a -> [Doc]
+prListRule_ (Rule _ _ items _) = for items $ \case
+  Right t       -> "render(" <> text (snd (renderCharOrString t)) <> ");"
+  Left c
+    | Just{} <- maybeTokenCat c
+                -> "visit" <> dat <> "(*i);"
+    | isList c  -> "iter" <> dat <> "(i+1, j);"
+    | otherwise -> "(*i)->accept(this);"
+    where
+    dat = text $ identCat $ normCat c
+
+-- This is the only part of the pretty printer that differs significantly
 -- between the versions with and without STL.
+-- The present version has been adapted from CFtoCPrinter.
 genPrintVisitorListNoStl :: (Cat, [Rule]) -> String
-genPrintVisitorListNoStl (cat@(ListCat c), rules) = unlines
-    [ "void PrintAbsyn::visit" ++ cl ++ "("++ cl ++ " *" ++ vname ++ ")"
+genPrintVisitorListNoStl (cat@(ListCat _), rules) = unlines $ concat
+  [ [ "void PrintAbsyn::visit" ++ cl ++ "("++ cl ++ " *" ++ vname ++ ")"
     , "{"
-    , "  while(" ++ vname +++ "!= 0)"
-    , "  {"
-    , "    if (" ++ vname ++ "->" ++ vname ++ "_ == 0)"
-    , "    {"
-    , visitMember
-    , optsep
-    , "      " ++ vname +++ "= 0;"
-    , "    }"
-    , "    else"
-    , "    {"
-    , visitMember
-    , render $ nest 6 $ renderListSepByPrecedence "_i_" renderSep separators
-    , "      " ++ vname +++ "=" +++ vname ++ "->" ++ vname ++ "_;"
-    , "    }"
-    , "  }"
-    , "}"
-    , ""
+      , "  if (" ++ vname +++ "== 0)"
+      , "  { /* nil */"
+      ]
+    , unlessNull (swRules isNilFun) $ \ docs ->
+      [ render $ nest 4 $ vcat docs ]
+    , [ "  }" ]
+    , unlessNull (swRules isOneFun) $ \ docs ->
+      [ "  else if (" ++ pre ++ vname ++ "_ == 0)"
+      , "  { /* last */"
+      , render $ nest 4 $ vcat docs
+      , "  }"
+      ]
+    , unlessNull (swRules isConsFun) $ \ docs ->
+      [ "  else"
+      , "  { /* cons */"
+      , render $ nest 4 $ vcat docs
+      , "  }"
+      ]
+    , [ "}"
+      , ""
+      ]
     ]
   where
-    visitMember = if isTokenCat c
-        then "      visit" ++ funName c ++ "(" ++ vname ++ "->" ++ member ++ ");"
-        else "      " ++ vname ++ "->" ++ member ++ "->accept(this);"
-    cl = identCat (normCat cat)
-    ecl = identCat (normCatOfList cat)
-    vname = map toLower cl
-    member = map toLower ecl ++ "_"
-    optsep = if hasOneFunc rules then "" else "      render(" ++ sep ++ ");"
-    sep = snd (renderCharOrString sep')
-    sep' = getCons rules
-    renderSep s = "render(" <> text (snd (renderCharOrString s)) <> ")"
-    separators = getSeparatorByPrecedence rules
+  cl          = identCat (normCat cat)
+  vname       = map toLower cl
+  pre         = vname ++ "->"
+  prules      = sortRulesByPrecedence rules
+  swRules f   = switchByPrecedence "_i_" $
+                  map (second $ sep . map text . prPrintRule_ pre) $
+                    uniqOn fst $ filter f prules
+                    -- Discard duplicates, can only handle one rule per precedence.
 genPrintVisitorListNoStl _ = error "genPrintVisitorListNoStl expects a ListCat"
 
 --Pretty Printer methods for a rule.
 prPrintRule :: Maybe String -> Rule -> String
-prPrintRule inPackage r@(Rule fun _ cats) | isProperLabel fun = unlines
-  [
-   "void PrintAbsyn::visit" ++ fun ++ "(" ++ fun +++ "*" ++ fnm ++ ")",
-   "{",
-   "  int oldi = _i_;",
-   lparen,
-   cats',
-   rparen,
-   "  _i_ = oldi;",
-   "}",
-   ""
+prPrintRule inPackage r@(Rule fun _ _ _) | isProperLabel fun = unlines $ concat
+  [ [ "void PrintAbsyn::visit" ++ funName fun ++ "(" ++ funName fun +++ "*" ++ fnm ++ ")"
+    , "{"
+    , "  int oldi = _i_;"
+    , parenCode "_L_PAREN"
+    , ""
+    ]
+  , prPrintRule_ (fnm ++ "->") r
+  , [ ""
+    , parenCode "_R_PAREN"
+    , "  _i_ = oldi;"
+    , "}"
+    , ""
+    ]
   ]
-   where
-    p = precRule r
-    (lparen, rparen) =
-      ("  if (oldi > " ++ show p ++ ") render(" ++ nsDefine inPackage "_L_PAREN" ++ ");\n",
-       "  if (oldi > " ++ show p ++ ") render(" ++ nsDefine inPackage "_R_PAREN" ++ ");\n")
-    cats' = concatMap (prPrintCat fnm) (numVars cats)
-    fnm = "p" --old names could cause conflicts
+  where
+  p = precRule r
+  parenCode x = "  if (oldi > " ++ show p ++ ") render(" ++ nsDefine inPackage x ++ ");"
+  fnm = "p" --old names could cause conflicts
 prPrintRule _ _ = ""
 
+prPrintRule_ :: IsFun a => String -> Rul a -> [String]
+prPrintRule_ pre (Rule _ _ items _) = map (prPrintItem pre) $ numVars items
+
 --This goes on to recurse to the instance variables.
-prPrintCat :: String -> Either (Cat, Doc) String -> String
-prPrintCat _ (Right t) = "  render(" ++ t' ++ ");\n"
-  where t' = snd (renderCharOrString t)
-prPrintCat fnm (Left (c, nt))
-  | isTokenCat c  = "  visit" ++ funName c ++ "(" ++ fnm ++ "->" ++ render nt ++ ");\n"
-  | isList c            = "  if(" ++ fnm ++ "->" ++ render nt ++ ") {" ++ accept ++ "}\n"
-  | otherwise           = "  " ++ accept ++ "\n"
+prPrintItem :: String -> Either (Cat, Doc) String -> String
+prPrintItem _   (Right t) = "  render(" ++ snd (renderCharOrString t) ++ ");"
+prPrintItem pre (Left (c, nt))
+  | Just t <- maybeTokenCat c
+              = "  visit" ++ t   ++ "(" ++ pre ++ s ++ ");"
+  | isList c  = "  " ++ setI (precCat c) ++
+                  "visit" ++ elt ++ "(" ++ pre ++ s ++ ");"
+  | otherwise = "  " ++ setI (precCat c) ++ pre ++ s ++ "->accept(this);"
   where
-    accept
-      | c == InternalCat =  "/* Internal Category */\n"
-      | otherwise        = setI (precCat c) ++ fnm ++ "->" ++ render nt ++ "->accept(this);"
+  s   = render nt
+  elt = identCat $ normCat c
 
 {- **** Abstract Syntax Tree Printer **** -}
 
@@ -577,20 +589,22 @@ prShowData False (cat@(ListCat c), _) =
     ecl = identCat (normCatOfList cat)
     vname = map toLower cl
     member = map toLower ecl ++ "_"
-    visitMember = if isTokenCat c
-      then "      visit" ++ funName c ++ "(" ++ vname ++ "->" ++ member ++ ");"
-      else "      " ++ vname ++ "->" ++ member ++ "->accept(this);"
+    visitMember
+      | Just t <- maybeTokenCat c =
+          "      visit" ++ t ++ "(" ++ vname ++ "->" ++ member ++ ");"
+      | otherwise =
+          "      " ++ vname ++ "->" ++ member ++ "->accept(this);"
 prShowData _ (cat, rules) =  --Not a list:
   abstract ++ concatMap prShowRule rules
   where
     cl = identCat (normCat cat)
-    abstract = case lookupRule (show cat) rules of
+    abstract = case lookupRule (noPosition $ catToStr cat) rules of
       Just _ -> ""
       Nothing ->  "void ShowAbsyn::visit" ++ cl ++ "(" ++ cl ++ " *p) {} //abstract class\n\n"
 
 --This prints all the methods for Abstract Syntax tree rules.
-prShowRule :: Rule -> String
-prShowRule (Rule fun _ cats) | isProperLabel fun = concat
+prShowRule :: IsFun f => Rul f -> String
+prShowRule (Rule f _ cats _) | isProperLabel f = concat
   [
    "void ShowAbsyn::visit" ++ fun ++ "(" ++ fun +++ "*" ++ fnm ++ ")\n",
    "{\n",
@@ -602,75 +616,62 @@ prShowRule (Rule fun _ cats) | isProperLabel fun = concat
    "}\n"
   ]
    where
-    (optspace, lparen, rparen) = if allTerms cats
-      then ("","","")
-      else ("  bufAppend(' ');\n", "  bufAppend('(');\n","  bufAppend(')');\n")
-    cats' = if allTerms cats
-        then ""
-        else concat (insertSpaces (map (prShowCat fnm) (numVars cats)))
+    fun = funName f
+    (optspace, lparen, rparen, cats')
+      | null [ () | Left _ <- cats ]  -- @all isRight cats@, but Data.Either.isRight requires base >= 4.7
+                  = ("", "", "", "")
+      | otherwise = ("  bufAppend(' ');\n", "  bufAppend('(');\n","  bufAppend(')');\n"
+                    , concat (insertSpaces (map (prShowCat fnm) (numVars cats))))
     insertSpaces [] = []
     insertSpaces (x:[]) = [x]
     insertSpaces (x:xs) = if x == ""
       then insertSpaces xs
       else x : "  bufAppend(' ');\n" : insertSpaces xs
-    allTerms [] = True
-    allTerms (Left _:_) = False
-    allTerms (_:zs) = allTerms zs
     fnm = "p" --other names could cause conflicts
 prShowRule _ = ""
 
 -- This recurses to the instance variables of a class.
 prShowCat :: String -> Either (Cat, Doc) String -> String
-prShowCat _ (Right _)               = ""
-prShowCat fnm (Left (cat,nt))
-  | isTokenCat cat              =
-    "  visit" ++ funName cat ++ "(" ++ fnm ++ "->" ++ render nt ++ ");\n"
-  | cat == InternalCat                = "/* Internal Category */\n"
-  | show (normCat $ strToCat $ render nt) /= render nt = accept
-  | otherwise                         =
-    concat [
-           "  bufAppend('[');\n",
-           "  if (" ++ fnm ++ "->" ++ render nt ++ ")" ++ accept,
-           "  bufAppend(']');\n"
-          ]
-  where accept = "  " ++ fnm ++ "->" ++ render nt ++ "->accept(this);\n"
+prShowCat _   (Right _) = ""
+prShowCat fnm (Left (cat, nt))
+  | Just t <- maybeTokenCat cat =
+      unlines
+        [ "  visit" ++ t ++ "(" ++ fnm ++ "->" ++ s ++ ");"
+        ]
+  | catToStr (normCat $ strToCat s) /= s =
+      unlines
+        [ accept
+        ]
+  | otherwise =
+      unlines
+        [ "  bufAppend('[');"
+        , "  if (" ++ fnm ++ "->" ++ s ++ ")" ++ accept
+        , "  bufAppend(']');"
+        ]
+  where
+  s = render nt
+  accept = "  " ++ fnm ++ "->" ++ s ++ "->accept(this);"
 
 {- **** Helper Functions Section **** -}
 
 -- from ListIdent to Ident
+baseName :: [a] -> [a]
 baseName = drop 4
 
-
---The visit-function name of a basic type
-funName :: Cat -> String
-funName (TokenCat c) | c `elem` builtin = c
-  where builtin = ["Integer", "Char", "String", "Double", "Ident" ]
-funName _ = "Ident" --User-defined type
---The visit-function name of a basic type
--- funName :: String -> String
--- funName v =
---     if "integer_" `isPrefixOf` v then "Integer"
---     else if "char_" `isPrefixOf` v then "Char"
---     else if "string_" `isPrefixOf` v then "String"
---     else if "double_" `isPrefixOf` v then "Double"
---     else if "ident_" `isPrefixOf` v then "Ident"
---     else "Ident" --User-defined type
-
 --Just sets the coercion level for parentheses in the Pretty Printer.
 setI :: Integer -> String
 setI n = "_i_ = " ++ show n ++ "; "
 
 --An extremely simple renderer for terminals.
 prRender :: Bool -> String
-prRender useStl = unlines
-  [
+prRender useStl = unlines $ concat
+  [ [
       "//You may wish to change render",
       "void PrintAbsyn::render(Char c)",
       "{",
       "  if (c == '{')",
       "  {",
-      "     bufAppend('\\n');",
-      "     indent();",
+      "     onEmptyLine();",
       "     bufAppend(c);",
       "     _n_ = _n_ + INDENT_WIDTH;",
       "     bufAppend('\\n');",
@@ -680,75 +681,104 @@ prRender useStl = unlines
       "     bufAppend(c);",
       "  else if (c == ')' || c == ']')",
       "  {",
-      "     backup();",
+      "     removeTrailingWhitespace();",
       "     bufAppend(c);",
+      "     bufAppend(' ');",
       "  }",
       "  else if (c == '}')",
       "  {",
-      "     int t;",
       "     _n_ = _n_ - INDENT_WIDTH;",
-      "     for (t=0; t<INDENT_WIDTH; t++) {",
-      "       backup();",
-      "     }",
+      "     onEmptyLine();",
       "     bufAppend(c);",
       "     bufAppend('\\n\');",
       "     indent();",
       "  }",
       "  else if (c == ',')",
       "  {",
-      "     backup();",
+      "     removeTrailingWhitespace();",
       "     bufAppend(c);",
       "     bufAppend(' ');",
       "  }",
       "  else if (c == ';')",
       "  {",
-      "     backup();",
+      "     removeTrailingWhitespace();",
       "     bufAppend(c);",
       "     bufAppend('\\n');",
       "     indent();",
       "  }",
+      "  else if (c == ' ') bufAppend(c);",
       "  else if (c == 0) return;",
       "  else",
       "  {",
-      "     bufAppend(' ');",
       "     bufAppend(c);",
       "     bufAppend(' ');",
       "  }",
       "}",
-      "",
-      let renderString = "void PrintAbsyn::render(String s_)" $$ codeblock 2
-                             [ "const char *s = s_.c_str() ;"
-                             , "if(strlen(s) > 0)"
-                             , codeblock 2
-                                 [ "bufAppend(s);"
-                                 , "bufAppend(' ');" ] ]
-      in if useStl then render renderString else "",
-      "void PrintAbsyn::render(const char *s)",
-      "{",
-      "  if(strlen(s) > 0)",
-      "  {",
-      "    bufAppend(s);",
-      "    bufAppend(' ');",
-      "  }",
-      "}",
-      "",
-      "void PrintAbsyn::indent()",
-      "{",
-      "  int n = _n_;",
-      "  while (n > 0)",
-      "  {",
-      "    bufAppend(' ');",
-      "    n--;",
-      "  }",
-      "}",
-      "",
-      "void PrintAbsyn::backup()",
-      "{",
-      "  if (buf_[cur_ - 1] == ' ')",
-      "  {",
-      "    buf_[cur_ - 1] = 0;",
-      "    cur_--;",
-      "  }",
-      "}",
       ""
+    ]
+  , when useStl
+    [ render $ vcat
+        [ "void PrintAbsyn::render(String s)"
+        , codeblock 2
+            [ "render(s.c_str());"
+            ]
+        , ""
+        ]
+    ]
+  , [ "bool allIsSpace(const char *s)"
+    , "{"
+    , "  char c;"
+    , "  while ((c = *s++))"
+    , "    if (! isspace(c)) return false;"
+    , "  return true;"
+    , "}"
+    , ""
+    ]
+  , [ "void PrintAbsyn::render(const char *s)"
+    , "{"
+    , "  if (*s) /* C string not empty */"
+    , "  {"
+    , "    if (allIsSpace(s)) {"
+    , "      backup();"
+    , "      bufAppend(s);"
+    , "    } else {"
+    , "      bufAppend(s);"
+    , "      bufAppend(' ');"
+    , "    }"
+    , "  }"
+    , "}"
+    , ""
+    , "void PrintAbsyn::indent()"
+    , "{"
+    , "  int n = _n_;"
+    , "  while (--n >= 0)"
+    , "    bufAppend(' ');"
+    , "}"
+    , ""
+    , "void PrintAbsyn::backup()"
+    , "{"
+    , "  if (cur_ && buf_[cur_ - 1] == ' ')"
+    , "    buf_[--cur_] = 0;"
+    , "}"
+    , ""
+    , "void PrintAbsyn::removeTrailingSpaces()"
+    , "{"
+    , "  while (cur_ && buf_[cur_ - 1] == ' ') --cur_;"
+    , "  buf_[cur_] = 0;"
+    , "}"
+    , ""
+    , "void PrintAbsyn::removeTrailingWhitespace()"
+    , "{"
+    , "  while (cur_ && (buf_[cur_ - 1] == ' ' || buf_[cur_ - 1] == '\\n')) --cur_;"
+    , "  buf_[cur_] = 0;"
+    , "}"
+    , ""
+    , "void PrintAbsyn::onEmptyLine()"
+    , "{"
+    , "  removeTrailingSpaces();"
+    , "  if (cur_ && buf_[cur_ - 1 ] != '\\n') bufAppend('\\n');"
+    , "  indent();"
+    , "}"
+    , ""
+    ]
   ]
