@@ -1,30 +1,18 @@
 {-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RecordWildCards #-}
 
 {-
     BNF Converter: Java Top File
     Copyright (C) 2004  Author:  Markus Forsberg, Peter Gammie,
                                  Michael Pellauer, Bjorn Bringert
 
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
 -}
 
 -------------------------------------------------------------------
 -- |
 -- Module      :  JavaTop
 -- Copyright   :  (C)opyright 2003, {markus, aarne, pellauer, peteg, bringert} at cs dot chalmers dot se
--- License     :  GPL (see COPYING for details)
 --
 -- Maintainer  :  {markus, aarne} at cs dot chalmers dot se
 -- Stability   :  alpha
@@ -37,9 +25,10 @@
 
 module BNFC.Backend.Java ( makeJava ) where
 
-import Prelude'
+import Prelude hiding ((<>))
 
-import System.FilePath (pathSeparator, isPathSeparator)
+import System.FilePath ((</>), (<.>), pathSeparator, isPathSeparator)
+import Data.Foldable (toList)
 import Data.List ( intersperse )
 
 import BNFC.Utils
@@ -68,15 +57,18 @@ import BNFC.PrettyPrint
 
 -- | This creates the Java files.
 makeJava :: SharedOptions -> CF -> MkFiles ()
-makeJava options@Options{..} cf = do
+makeJava opt = makeJava' opt{ lang = mkName javaReserved SnakeCase $ lang opt }
+  -- issue #212: make a legal package name, see also
+  -- https://docs.oracle.com/javase/tutorial/java/package/namingpkgs.html
+
+makeJava' :: SharedOptions -> CF -> MkFiles ()
+makeJava' options@Options{..} cf = do
      -- Create the package directories if necessary.
-    let packageBase  = case inPackage of
-                           Nothing -> lang
-                           Just p -> p ++ "." ++ lang
-        packageAbsyn = packageBase ++ "." ++ "Absyn"
+    let packageBase  = maybe id (+.+) inPackage lang
+        packageAbsyn = packageBase +.+ "Absyn"
         dirBase      = pkgToDir packageBase
         dirAbsyn     = pkgToDir packageAbsyn
-        javaex str   = dirBase ++ str +.+ "java"
+        javaex str   = dirBase </> str <.> "java"
         bnfcfiles    =
           bnfcVisitorsAndTests
             packageBase
@@ -92,15 +84,12 @@ makeJava options@Options{..} cf = do
                 (head $ results lexmake) -- lexer class
                 (head $ results parmake) -- parser class
             )
-        makebnfcfile x = mkfile (javaex (fst $ x bnfcfiles))
+        makebnfcfile x = mkfile (javaex (fst $ x bnfcfiles)) comment
                                         (snd $ x bnfcfiles)
 
-    let absynFiles = remDups $ cf2JavaAbs packageBase packageAbsyn cf rp
-        absynBaseNames = map fst absynFiles
-        absynFileNames = map (dirAbsyn ++) absynBaseNames
-    let writeAbsyn (filename, contents) =
-          mkfile (dirAbsyn ++ filename ++ ".java") contents
-    mapM_ writeAbsyn absynFiles
+    let absynFiles = remDups $ cf2JavaAbs dirAbsyn packageBase packageAbsyn cf rp
+        absynFileNames = map fst absynFiles
+    mapM_ (\ (n, s) -> mkfile (n <.> "java") comment s) absynFiles
     makebnfcfile bprettyprinter
     makebnfcfile bskel
     makebnfcfile bcompos
@@ -110,18 +99,18 @@ makeJava options@Options{..} cf = do
     makebnfcfile btest
     let (lex, env) = lexfun packageBase cf
     -- Where the lexer file is created. lex is the content!
-    mkfile (dirBase ++ inputfile lexmake ) lex
-    liftIO $ putStrLn $ "   (Tested with "+++ toolname lexmake
-                                          +++ toolversion lexmake  +++")"
+    mkfile (dirBase </> inputfile lexmake ) commentWithEmacsModeHint lex
+    liftIO $ putStrLn $ "   (Tested with" +++ toolname lexmake
+                                          +++ toolversion lexmake  ++ ")"
     -- where the parser file is created.
-    mkfile (dirBase ++ inputfile parmake)
+    mkfile (dirBase </> inputfile parmake) commentWithEmacsModeHint
           $ parsefun packageBase packageAbsyn cf rp env
     liftIO $ putStrLn $
       if supportsEntryPoints parmake
        then "(Parser created for all categories)"
-       else "   (Parser created only for category " ++ show (firstEntry cf) ++ ")"
-    liftIO $ putStrLn $ "   (Tested with " +++ toolname parmake
-                                           +++ toolversion parmake +++ ")"
+       else "   (Parser created only for category " ++ prettyShow (firstEntry cf) ++ ")"
+    liftIO $ putStrLn $ "   (Tested with"  +++ toolname parmake
+                                           +++ toolversion parmake ++ ")"
     Makefile.mkMakefile options $
         makefile dirBase dirAbsyn absynFileNames parselexspec
   where
@@ -130,7 +119,7 @@ makeJava options@Options{..} cf = do
                            Just {} -> remDups as
                            Nothing -> (a, b) : remDups as
     pkgToDir :: String -> FilePath
-    pkgToDir s = replace '.' pathSeparator s ++ [pathSeparator]
+    pkgToDir = replace '.' pathSeparator
 
     parselexspec = parserLexerSelector lang javaLexerParser rp
     lexfun       = cf2lex $ lexer parselexspec
@@ -138,6 +127,7 @@ makeJava options@Options{..} cf = do
     parmake      = makeparserdetails (parser parselexspec)
     lexmake      = makelexerdetails  (lexer parselexspec)
     rp           = (Options.linenumbers options)
+    commentWithEmacsModeHint = comment . ("-*- Java -*- " ++)
 
 makefile ::  FilePath -> FilePath -> [String] -> ParserLexerSpecification -> String -> Doc
 makefile  dirBase dirAbsyn absynFileNames jlexpar basename = vcat $
@@ -162,41 +152,41 @@ makefile  dirBase dirAbsyn absynFileNames jlexpar basename = vcat $
                 ]++
     [-- running the lexergen: output of lexer -> input of lexer : calls lexer
     let ff = filename lexmake -- name of input file without extension
-        dirBaseff = dirBase ++ ff -- prepend directory
-        inp = dirBase ++ inputfile lexmake in
-        Makefile.mkRule (dirBaseff +.+ "java") [ inp ]
+        dirBaseff = dirBase </> ff -- prepend directory
+        inp = dirBase </> inputfile lexmake in
+        Makefile.mkRule (dirBaseff <.> "java") [ inp ]
         [ "${LEXER} ${LEXER_FLAGS} "++ inp ]
 
     -- running the parsergen, these there are its outputs
     -- output of parser -> input of parser : calls parser
-  , let inp = dirBase ++ inputfile parmake in
-        Makefile.mkRule (unwords (map (dirBase++) (dotJava $ results parmake)))
+  , let inp = dirBase </> inputfile parmake in
+        Makefile.mkRule (unwords (map (dirBase </>) (dotJava $ results parmake)))
           [ inp ] $
           ("${PARSER} ${PARSER_FLAGS} " ++ inp) :
-          ["mv " ++ unwords (dotJava $ results parmake) +++ dirBase
+          ["mv " ++ unwords (dotJava $ results parmake) +++ dirBase ++ [pathSeparator]
               | moveresults parmake]
   -- Class of the output of lexer generator wants java of :
   -- output of lexer and parser generator
-  , let lexerOutClass = dirBase ++ filename lexmake +.+ "class"
-        outname x = dirBase ++ x +.+ "java"
+  , let lexerOutClass = dirBase </> filename lexmake <.> "class"
+        outname x = dirBase </> x <.> "java"
         deps = map outname (results lexmake ++ results parmake) in
         Makefile.mkRule lexerOutClass deps []
     ]++
   reverse [Makefile.mkRule tar dep [] |
     (tar,dep) <- partialParserGoals dirBase (results parmake)]
-  ++[ Makefile.mkRule (dirBase ++ "PrettyPrinter.class")
-        [ dirBase ++ "PrettyPrinter.java" ] []
+  ++[ Makefile.mkRule (dirBase </> "PrettyPrinter.class")
+        [ dirBase </> "PrettyPrinter.java" ] []
     -- Removes all the class files created anywhere
-    , Makefile.mkRule "clean" [] [ "rm -f " ++ dirAbsyn ++ "*.class" ++ " "
-                                            ++ dirBase ++ "*.class" ]
+    , Makefile.mkRule "clean" [] [ "rm -f " ++ dirAbsyn </> "*.class" ++ " "
+                                            ++ dirBase </> "*.class" ]
     -- Remains the same
     , Makefile.mkRule "distclean" [ "vclean" ] []
     -- removes everything
     , Makefile.mkRule "vclean" []
         [ " rm -f " ++ absynJavaSrc ++ " " ++ absynJavaClass
-        , " rm -f " ++ dirAbsyn ++ "*.class"
+        , " rm -f " ++ dirAbsyn </> "*.class"
         , " rmdir " ++ dirAbsyn
-        , " rm -f " ++ unwords (map (dirBase ++) $
+        , " rm -f " ++ unwords (map (dirBase </>) $
                     [ inputfile lexmake
                     , inputfile parmake
                     ]
@@ -225,8 +215,9 @@ makefile  dirBase dirAbsyn absynFileNames jlexpar basename = vcat $
       lexmake           = makelexerdetails (lexer jlexpar)
       absynJavaSrc      = unwords (dotJava absynFileNames)
       absynJavaClass    = unwords (dotClass absynFileNames)
-      classes = prependPath dirBase lst
+      classes = map (dirBase </>) lst
       lst = dotClass (results lexmake) ++ [ "PrettyPrinter.class", "Test.class"
+          , "VisitSkel.class"
           , "ComposVisitor.class", "AbstractVisitor.class"
           , "FoldVisitor.class", "AllVisitor.class"]++
            dotClass (results parmake) ++ ["Test.class"]
@@ -307,15 +298,11 @@ antlrtest = javaTest $ JavaTestParams
   , jtpInvocation         =
       \ pbase pabs dat enti -> vcat
          [
-           let rulename = getRuleName (show enti)
+           let rulename = getRuleName $ startSymbol $ render enti
                typename = text rulename
                methodname = text $ firstLowerCase rulename
            in
                pbase <> "." <> typename <> "Context pc = p." <> methodname <> "();"
-         , "org.antlr.v4.runtime.Token _tkn = p.getInputStream().getTokenSource().nextToken();"
-         , "if(_tkn.getType() != -1) throw new TestError"
-                 <> "(\"Stream does not end with EOF\","
-                 <> "_tkn.getLine(),_tkn.getCharPositionInLine());"
          , pabs <> "." <> dat <+> "ast = pc.result;"
          ]
   , jtpErrMsg             =
@@ -460,7 +447,7 @@ jlexmakedetails = MakeDetails
     , filename            = "Yylex"
     , fileextension       = ""
     , toolname            = "JLex"
-    , toolversion         = "1.2.6."
+    , toolversion         = "1.2.6"
     , supportsEntryPoints = False
     , results             = ["Yylex"]
     , other_results       = []
@@ -470,7 +457,7 @@ jlexmakedetails = MakeDetails
 jflexmakedetails = jlexmakedetails
     { executable  = "jflex"
     , toolname    = "JFlex"
-    , toolversion = "1.4.3"
+    , toolversion = "1.4.3 - 1.7.0"
     }
 
 cupmakedetails rp = MakeDetails
@@ -492,9 +479,8 @@ cupmakedetails rp = MakeDetails
 antlrmakedetails :: String -> MakeFileDetails
 antlrmakedetails l = MakeDetails
     { executable = runJava "org.antlr.v4.Tool"
-    , flags               = \x -> unwords $
-                                    let path    = take (length x - 1) x
-                                        pointed = map cnv path
+    , flags               = \ path -> unwords $
+                                    let pointed = map cnv path
                                         cnv y   = if isPathSeparator y
                                                         then '.'
                                                         else y
@@ -503,20 +489,21 @@ antlrmakedetails l = MakeDetails
     , filename            = l
     , fileextension       = "g4"
     , toolname            = "ANTLRv4"
-    , toolversion         = "4.5.1"
+    , toolversion         = "4.9"
     , supportsEntryPoints = True
     , results             = [l]
-    , other_results       = map (l ++) [".tokens","BaseListener.java","Listener.java"]
+    , other_results       = map (l ++)
+        [ ".interp"              -- added after ANTLR 4.5
+        , ".tokens"
+        , "BaseListener.java"
+        ,"Listener.java"
+        ]
     , moveresults         = False
     }
 
-prependPath , appendExtension :: String -> [String] -> [String]
-prependPath s fi     = [ s ++ x  | x <- fi ]
-appendExtension s fi = [ x +.+ s | x <- fi ]
-
-dotJava,dotClass :: [String] -> [String]
-dotJava  = appendExtension "java"
-dotClass = appendExtension "class"
+dotJava, dotClass :: [String] -> [String]
+dotJava  = map (<.> "java")
+dotClass = map (<.> "class")
 
 type CFToJava = String -> String -> CF -> String
 
@@ -548,16 +535,17 @@ bnfcVisitorsAndTests pbase pabsyn cf cf0 cf1 cf2 cf3 cf4 cf5 cf6 =
     }
   where app x = x pbase pabsyn cf
 
-inputfile x = filename x ++ case fileextension x of
-                                "" -> ""
-                                a -> '.':a
+inputfile :: MakeFileDetails -> String
+inputfile x
+  | null (fileextension x) = filename x
+  | otherwise              = filename x <.> fileextension x
 
 -- |  constructs the rules regarding the parser in the makefile
 partialParserGoals :: String -> [String] -> [(String, [String])]
 partialParserGoals _ []          = []
-partialParserGoals dbas (x:rest) =
-    (dbas ++ x +.+ "class", map (\ y -> dbas ++ y +.+ "java") (x:rest))
-        : partialParserGoals dbas rest
+partialParserGoals dirBase (x:rest) =
+    (dirBase </> x <.> "class", map (\ y -> dirBase </> y <.> "java") (x:rest))
+        : partialParserGoals dirBase rest
 
 -- | Creates the Test.java class.
 javaTest :: JavaTestParams -> TestClass
@@ -577,11 +565,12 @@ javaTest (JavaTestParams
     cf =
     render $ vcat $ concat $
       [ [ "package" <+> text packageBase <> ";"
+        , ""
         , "import" <+> text packageBase <> ".*;"
-        , "import" <+> text packageAbsyn <> ".*;"
         , "import java.io.*;"
         ]
       , map importfun imports
+      , [ "" ]
       , errhand err
       , [ ""
         , "public class Test"
@@ -594,14 +583,13 @@ javaTest (JavaTestParams
                 [ "try"
                 , codeblock 2
                     [ "Reader input;"
-                    , "if (args.length == 0)"
-                       <> "input = new InputStreamReader(System.in);"
+                    , "if (args.length == 0) input = new InputStreamReader(System.in);"
                     , "else input = new FileReader(args[0]);"
-                    , "l = new "<>lexerconstruction lx "(input)"
+                    , "l = new " <> lexerconstruction lx "(input)"
                     ]
                 , "catch(IOException e)"
-                , codeblock 2 [ "System.err.println"
-                        <>"(\"Error: File not found: \" + args[0]);"
+                , codeblock 2
+                    [ "System.err.println(\"Error: File not found: \" + args[0]);"
                     , "System.exit(1);"
                     ]
                 , "p = new "<> parserconstruction px "l"
@@ -609,26 +597,32 @@ javaTest (JavaTestParams
             , ""
             , "public" <+> text packageAbsyn <> "." <> dat
                 <+> "parse() throws Exception"
-            , codeblock 2
-                [ "/* The default parser is the first-defined entry point. */"
-                , "/* Other options are: */"
-                , "/* " <> fsep (punctuate "," (showOpts (tail eps))) <> " */"
-                , invocation px (text packageAbsyn) dat absentity
-                , printOuts [ "\"Parse Succesful!\""
-                    , "\"[Abstract Syntax]\""
-                    , "PrettyPrinter.show(ast)"
-                    , "\"[Linearized Tree]\""
-                    , "PrettyPrinter.print(ast)"
-                    ]
-                , "return ast;"
+            , codeblock 2 $ concat
+                [ [ "/* The default parser is the first-defined entry point. */" ]
+                , unlessNull (drop 1 eps) $ \ eps' ->
+                  [ "/* Other options are: */"
+                  , "/* " <> fsep (punctuate "," (showOpts eps')) <> " */"
+                  ]
+                , [ invocation px (text packageAbsyn) dat absentity
+                  , printOuts
+                     [ "\"Parse Succesful!\""
+                     , "\"[Abstract Syntax]\""
+                     , "PrettyPrinter.show(ast)"
+                     , "\"[Linearized Tree]\""
+                     , "PrettyPrinter.print(ast)"
+                     ]
+                  , "return ast;"
+                  ]
                 ]
             , ""
             , "public static void main(String args[]) throws Exception"
-            , codeblock 2 [ "Test t = new Test(args);"
+            , codeblock 2
+                [ "Test t = new Test(args);"
                 , "try"
                 , codeblock 2 [ "t.parse();" ]
-                ,"catch("<>text err<+>"e)"
-                , codeblock 2 [ "System.err.println(\""<>text errmsg<>"\");"
+                ,"catch(" <> text err <+> "e)"
+                , codeblock 2
+                    [ "System.err.println(\"" <> text errmsg <> "\");"
                     , "System.err.println(\"     \" + e.getMessage());"
                     , "System.exit(1);"
                     ]
@@ -645,7 +639,7 @@ javaTest (JavaTestParams
       px             = text parser
       dat            = text $ identCat $ normCat def  -- Use for AST types.
       absentity      = text $ identCat def            -- Use for parser/printer name.
-      eps            = allEntryPoints cf
+      eps            = toList $ allEntryPoints cf
       def            = head eps
 
 -- | Error handling in ANTLR.
