@@ -1,29 +1,10 @@
-{-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE OverloadedStrings #-}
 
 {-
     BNF Converter: Java Pretty Printer generator
     Copyright (C) 2004  Author:  Michael Pellauer, Bjorn Bringert
 
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
--}
-
-
-{-
-   **************************************************************
-    BNF Converter Module
-
     Description   : This module generates the Java Pretty Printer
                     class. In addition, since there's no good way
                     to display a class heirarchy (toString() doesn't
@@ -37,29 +18,29 @@
     Author        : Michael Pellauer (pellauer@cs.chalmers.se),
                     Bjorn Bringert (bringert@cs.chalmers.se)
 
-    License       : GPL (GNU General Public License)
-
     Created       : 24 April, 2003
 
     Modified      : 9 Aug, 2004
-
     Added string buffer for efficiency (Michael, August 03)
-   **************************************************************
+
 -}
+
 module BNFC.Backend.Java.CFtoJavaPrinter15 ( cf2JavaPrinter ) where
 
-import Prelude'
+import Prelude hiding ((<>))
 
-import BNFC.Backend.Java.CFtoJavaAbs15
+import Data.Bifunctor ( second )
+import Data.Char      ( toLower, isSpace )
+import Data.Either    ( lefts )
+import Data.List      ( intersperse )
 
 import BNFC.CF
-import BNFC.Backend.Common (renderListSepByPrecedence)
-import BNFC.Backend.Common.NamedVariables
-import BNFC.Utils ( (+++) )
-import Data.List
-import Data.Char ( toLower, isSpace )
-import Data.Either (lefts)
 import BNFC.PrettyPrint
+import BNFC.Utils     ( (+++), for, unless, unlessNull, uniqOn )
+
+import BNFC.Backend.Common ( switchByPrecedence )
+import BNFC.Backend.Common.NamedVariables
+import BNFC.Backend.Java.CFtoJavaAbs15
 
 --Produces the PrettyPrinter class.
 --It will generate two methods "print" and "show"
@@ -82,7 +63,6 @@ cf2JavaPrinter packageBase packageAbsyn cf =
     groups = fixCoercions (ruleGroupsInternals cf)
     header = unlines [
       "package" +++ packageBase ++ ";",
-      "import" +++ packageAbsyn ++ ".*;",
       "",
       "public class PrettyPrinter",
       "{",
@@ -96,36 +76,69 @@ cf2JavaPrinter packageBase packageAbsyn cf =
       ]
     footer = unlines [ --later only include used categories
       "  private static void pp(Integer n, int _i_) { buf_.append(n); buf_.append(\" \"); }",
-      "  private static void pp(Double d, int _i_) { buf_.append(d); buf_.append(\" \"); }",
+      "  private static void pp(Double d, int _i_) { buf_.append(String.format(java.util.Locale.ROOT, \"%.15g \", d)); }",
       "  private static void pp(String s, int _i_) { buf_.append(s); buf_.append(\" \"); }",
       "  private static void pp(Character c, int _i_) { buf_.append(\"'\" + c.toString() + \"'\"); buf_.append(\" \"); }",
       "  private static void sh(Integer n) { render(n.toString()); }",
-      "  private static void sh(Double d) { render(d.toString()); }",
-      "  private static void sh(Character c) { render(c.toString()); }",
+      "  private static void sh(Double d) { render(String.format(java.util.Locale.ROOT, \"%.15g\", d)); }",
+      "  private static void sh(Character c) { render(\"'\" + c.toString() + \"'\"); }",
       "  private static void sh(String s) { printQuoted(s); }",
+      "",
       "  private static void printQuoted(String s) { render(\"\\\"\" + s + \"\\\"\"); }",
+      "",
       "  private static void indent()",
       "  {",
       "    int n = _n_;",
       "    while (n > 0)",
       "    {",
-      "      buf_.append(\" \");",
+      "      buf_.append(\' \');",
       "      n--;",
       "    }",
       "  }",
+      "",
       "  private static void backup()",
       "  {",
-      "     if (buf_.charAt(buf_.length() - 1) == ' ') {",
-      "      buf_.setLength(buf_.length() - 1);",
-      "    }",
+      "    int prev = buf_.length() - 1;",
+      "    if (prev >= 0 && buf_.charAt(prev) == ' ')",
+      "      buf_.setLength(prev);",
       "  }",
+      "",
       "  private static void trim()",
       "  {",
-      "     while (buf_.length() > 0 && buf_.charAt(0) == ' ')",
-      "        buf_.deleteCharAt(0); ",
-      "    while (buf_.length() > 0 && buf_.charAt(buf_.length()-1) == ' ')",
-      "        buf_.deleteCharAt(buf_.length()-1);",
+      "    // Trim initial spaces",
+      "    int end = 0;",
+      "    int len = buf_.length();",
+      "    while (end < len && buf_.charAt(end) == ' ')",
+      "      end++; ",
+      "    buf_.delete(0, end);",
+      "    // Trim trailing spaces",
+      "    removeTrailingSpaces();",
       "  }",
+      "",
+      "  private static void removeTrailingSpaces()",
+      "  {",
+      "    int end = buf_.length();",
+      "    while (end > 0 && buf_.charAt(end-1) == ' ')",
+      "      end--;",
+      "    buf_.setLength(end);",
+      "  }",
+      "",
+      "  private static void removeTrailingWhitespace()",
+      "  {",
+      "    int end = buf_.length();",
+      "    while (end > 0 && (buf_.charAt(end-1) == ' ' || buf_.charAt(end-1) == '\\n'))",
+      "      end--;",
+      "    buf_.setLength(end);",
+      "  }",
+      "",
+      "  private static void onEmptyLine()",
+      "  {",
+      "    removeTrailingSpaces();",
+      "    int len = buf_.length();",
+      "    if (len > 0 && buf_.charAt(len-1) != '\\n') buf_.append(\"\\n\");",
+      "    indent();",
+      "  }",
+      "",
       "  private static int _n_ = 0;",
       "  private static StringBuilder buf_ = new StringBuilder(INITIAL_BUFFER_SIZE);",
       "}"
@@ -140,8 +153,7 @@ prRender = unlines
       "  {",
       "    if (s.equals(\"{\"))",
       "    {",
-      "       buf_.append(\"\\n\");",
-      "       indent();",
+      "       onEmptyLine();",
       "       buf_.append(s);",
       "       _n_ = _n_ + INDENT_WIDTH;",
       "       buf_.append(\"\\n\");",
@@ -151,35 +163,37 @@ prRender = unlines
       "       buf_.append(s);",
       "    else if (s.equals(\")\") || s.equals(\"]\"))",
       "    {",
-      "       backup();",
+      "       removeTrailingWhitespace();",
       "       buf_.append(s);",
       "       buf_.append(\" \");",
       "    }",
       "    else if (s.equals(\"}\"))",
       "    {",
-      "       int t;",
       "       _n_ = _n_ - INDENT_WIDTH;",
-      "       for(t=0; t<INDENT_WIDTH; t++) {",
-      "         backup();",
-      "       }",
+      "       onEmptyLine();",
       "       buf_.append(s);",
       "       buf_.append(\"\\n\");",
       "       indent();",
       "    }",
       "    else if (s.equals(\",\"))",
       "    {",
-      "       backup();",
+      "       removeTrailingWhitespace();",
       "       buf_.append(s);",
       "       buf_.append(\" \");",
       "    }",
       "    else if (s.equals(\";\"))",
       "    {",
-      "       backup();",
+      "       removeTrailingWhitespace();",
       "       buf_.append(s);",
       "       buf_.append(\"\\n\");",
       "       indent();",
       "    }",
       "    else if (s.equals(\"\")) return;",
+      "    else if (s.trim().equals(\"\"))",
+      "    {",
+      "       backup();",
+      "       buf_.append(s);",
+      "    }",
       "    else",
       "    {",
       "       buf_.append(s);",
@@ -216,28 +230,37 @@ prEntryPoints packageAbsyn cf =
     cat' = identCat cat
 
 prData :: String ->  [UserDef] -> (Cat, [Rule]) -> String
-prData packageAbsyn user (cat, rules) = unlines k
-    where
-      k = if isList cat
-           then
-           ["  private static void pp(" ++ packageAbsyn ++ "."
-                ++ identCat (normCat cat) +++ "foo, int _i_)"
-            , "  {"
-            , render $ nest 5 $ prList user cat rules <> "  }"
-           ]
-           else --not a list
-           [
-            "  private static void pp(" ++ packageAbsyn ++ "."
-                ++ identCat (normCat cat) +++ "foo, int _i_)",
-            "  {",
-            concat (addElse $ map (prRule packageAbsyn) rules) ++ "  }"
-           ]
-      addElse = map ("    "++). intersperse "else " . filter (not . null)
-        . map (dropWhile isSpace)
+prData packageAbsyn user (cat, rules)
+  | isList cat = unlines $ concat
+      [ [ "  private static void pp(" ++ packageAbsyn ++ "." ++ dat ++ " foo, int _i_)"
+        , "  {"
+        , "    pp" ++ dat ++ "(foo.iterator(), _i_);"
+        , "  }"
+        , ""
+        , "  private static void pp" ++ dat ++ "(java.util.Iterator<" ++ et ++ "> it, int _i_)"
+        , "  {"
+        ]
+      , map ("    " ++) $ prList dat et rules
+      , [ "  }"
+        , ""
+        ]
+      ]
+  | otherwise = unlines $
+      [ "  private static void pp(" ++ packageAbsyn ++ "."
+           ++ dat +++ "foo, int _i_)"
+      , "  {"
+      , concat (addElse $ map (prRule packageAbsyn) rules)
+      , "  }"
+      ]
+  where
+  dat = identCat (normCat cat)
+  et  = typename packageAbsyn user $ identCat $ normCatOfList cat
+  addElse = map ("    " ++) . intersperse "else " . filter (not . null)
+    . map (dropWhile isSpace)
 
 
 prRule :: String -> Rule -> String
-prRule packageAbsyn r@(Rule fun _c cats) | not (isCoercion fun || isDefinedRule fun) = concat
+prRule packageAbsyn r@(Rule f _c cats _) | not (isCoercion f || isDefinedRule f) = concat
     [ "    if (foo instanceof" +++ packageAbsyn ++ "." ++ fun ++ ")\n"
     , "    {\n"
     , "       " ++ packageAbsyn ++ "." ++ fun +++ fnm +++ "= ("
@@ -248,67 +271,82 @@ prRule packageAbsyn r@(Rule fun _c cats) | not (isCoercion fun || isDefinedRule
     , "    }\n"
     ]
   where
+    fun = funName f
     p = precRule r
     (lparen, rparen) =
         ("       if (_i_ > " ++ show p ++ ") render(_L_PAREN);\n",
         "       if (_i_ > " ++ show p ++ ") render(_R_PAREN);\n")
     cats' = case cats of
         [] -> ""
-        _  -> concatMap (render . prCat (text fnm)) (numVars cats)
+        _  -> concatMap (render . prItem (text fnm)) (numVars cats)
     fnm = '_' : map toLower fun
 
 prRule _nm _ = ""
 
--- |
---
--- >>> let lfoo = ListCat (Cat "Foo")
--- >>> prList [] lfoo [Rule "[]" lfoo [], Rule "(:)" lfoo [Left (Cat "Foo"), Right ".", Left lfoo]]
--- for (java.util.Iterator<Foo> it = foo.iterator(); it.hasNext();)
--- {
---   pp(it.next(), _i_);
---   if (it.hasNext()) {
---     render(".");
---   } else {
---     render(".");
---   }
--- }
+prList :: String -> String -> [Rule] -> [String]
+prList dat et rules = concat
+    [ if null docs0 then
+      [ "if (it.hasNext())" ]
+      else
+      [ "if (!it.hasNext())"
+      , "{ /* nil */"
+      , render $ nest 4 $ vcat docs0
+      , "}"
+      , "else"
+      ]
+    , if null docs1 then
+      [ "{ /* cons */"
+      , "  " ++ et ++ " el = it.next();"
+      ]
+      else
+      [ "{"
+      , "  " ++ et ++ " el = it.next();"
+      , "  if (!it.hasNext())"
+      , "  { /* last */"
+      , render $ nest 4 $ vcat docs1
+      , "  }"
+      , "  else"
+      , "  { /* cons */"
+      ]
+    , unlessNull (swRules isConsFun) $ \ docs ->
+      [ render $ nest (if null docs1 then 2 else 4) $ vcat docs
+      ]
+    , unless (null docs1) [ "  }" ]
+    , [ "}" ]
+    ]
+  where
+  prules      = sortRulesByPrecedence rules
+  swRules f   = switchByPrecedence "_i_" $
+                  map (second $ sep . map text . prListRule_ dat) $
+                    uniqOn fst $ filter f prules
+                    -- Discard duplicates, can only handle one rule per precedence.
+  docs0       = swRules isNilFun
+  docs1       = swRules isOneFun
 
-prList :: [UserDef] -> Cat -> [Rule] -> Doc
-prList user c rules =
-    "for (java.util.Iterator<" <> et <> "> it = foo.iterator(); it.hasNext();)"
-    $$ codeblock 2
-        [ "pp(it.next(), _i_);"
-        , "if (it.hasNext()) {"
-        , nest 2 (renderListSepByPrecedence "_i_" renderSep
-            (getSeparatorByPrecedence rules))
-        , "} else {"
-        , nest 2 (renderSep optsep <> ";")
-        , "}"
-        ]
-   where
-    et = text $ cat2JavaType user $ normCatOfList c
-    sep = escapeChars $ getCons rules
-    optsep = if hasOneFunc rules then "" else sep
-    renderSep x = "render(\"" <> text x <>"\")"
+-- | Only render the rhs (items) of a list rule.
+
+prListRule_ :: IsFun a => String -> Rul a -> [String]
+prListRule_ dat (Rule _ _ items _) = for items $ \case
+  Right t                  -> "render(\"" ++ escapeChars t ++ "\");"
+  Left (TokenCat "String") -> "printQuoted(el);"
+  Left (ListCat _)         -> "pp" ++ dat ++ "(it, _i_);"
+  Left _                   -> "pp(el, _i_);"
 
 -- |
--- >>> prCat "F" (Right "++")
+-- >>> prItem "F" (Right "++")
 --        render("++");
 -- <BLANKLINE>
--- >>> prCat "F" (Left (TokenCat "String", "string_"))
+-- >>> prItem "F" (Left (TokenCat "String", "string_"))
 --        printQuoted(F.string_);
 -- <BLANKLINE>
--- >>> prCat "F" (Left (InternalCat, "#_"))
--- <BLANKLINE>
--- >>> prCat "F" (Left (Cat "Abc", "abc_"))
+-- >>> prItem "F" (Left (Cat "Abc", "abc_"))
 --        pp(F.abc_, 0);
 -- <BLANKLINE>
-prCat :: Doc -> Either (Cat, Doc) String -> Doc
-prCat _ (Right t) = nest 7 ("render(\"" <> text(escapeChars t) <> "\");\n")
-prCat fnm (Left (TokenCat "String", nt))
+prItem :: Doc -> Either (Cat, Doc) String -> Doc
+prItem _ (Right t) = nest 7 ("render(\"" <> text(escapeChars t) <> "\");\n")
+prItem fnm (Left (TokenCat "String", nt))
     = nest 7 ("printQuoted(" <> fnm <> "." <> nt <> ");\n")
-prCat _ (Left (InternalCat, _)) = empty
-prCat fnm (Left (cat, nt))
+prItem fnm (Left (cat, nt))
     = nest 7 ("pp(" <> fnm <> "." <> nt <> ", " <> integer (precCat cat) <> ");\n")
 
 --The following methods generate the Show function.
@@ -321,7 +359,7 @@ shData packageAbsyn user (cat, rules) = unlines k
           [ "  private static void sh(" ++ packageAbsyn ++ "."
                 ++ identCat (normCat cat) +++ "foo)"
           , "  {"
-          , shList user cat rules ++ "  }"
+          , shList packageAbsyn user cat rules ++ "  }"
           ]
           else
           [ "  private static void sh(" ++ packageAbsyn ++ "."
@@ -332,7 +370,7 @@ shData packageAbsyn user (cat, rules) = unlines k
 
 
 shRule :: String -> Rule -> String
-shRule packageAbsyn (Rule fun _c cats) | not (isCoercion fun || isDefinedRule fun) = unlines
+shRule packageAbsyn (Rule f _c cats _) | not (isCoercion f || isDefinedRule f) = unlines
     [ "    if (foo instanceof" +++ packageAbsyn ++ "." ++ fun ++ ")"
     , "    {"
     , "       " ++ packageAbsyn ++ "." ++ fun +++ fnm +++ "= ("
@@ -340,6 +378,7 @@ shRule packageAbsyn (Rule fun _c cats) | not (isCoercion fun || isDefinedRule fu
     , members ++ "    }"
     ]
   where
+    fun = funName f
     members = concat [ lparen
                      , "       render(\"" ++ escapeChars fun ++ "\");\n"
                      , cats'
@@ -356,8 +395,8 @@ shRule packageAbsyn (Rule fun _c cats) | not (isCoercion fun || isDefinedRule fu
     fnm = '_' : map toLower fun
 shRule _nm _ = ""
 
-shList :: [UserDef] -> Cat -> [Rule] -> String
-shList user c _rules = unlines
+shList :: String -> [UserDef] -> Cat -> [Rule] -> String
+shList packageAbsyn user c _rules = unlines
   [
    "     for (java.util.Iterator<" ++ et
           ++ "> it = foo.iterator(); it.hasNext();)",
@@ -368,7 +407,7 @@ shList user c _rules = unlines
    "     }"
   ]
     where
-    et = cat2JavaType user $ normCatOfList c
+    et = typename packageAbsyn user $ identCat $ normCatOfList c
 
 -- |
 -- >>> shCat "F" (ListCat (Cat "A"), "lista_")
@@ -376,8 +415,6 @@ shList user c _rules = unlines
 --        sh(F.lista_);
 --        render("]");
 -- <BLANKLINE>
--- >>> shCat "F" (InternalCat, "#_")
--- <BLANKLINE>
 -- >>> shCat "F" (Cat "A", "a_")
 --        sh(F.a_);
 -- <BLANKLINE>
@@ -386,7 +423,6 @@ shCat fnm (ListCat _, vname) = vcat
     [ "       render(\"[\");"
     , "       sh(" <> fnm <> "." <> vname <> ");"
     , "       render(\"]\");\n" ]
-shCat _ (InternalCat, _)     = empty
 shCat fname (_, vname)       = "       sh(" <> fname <> "." <> vname <> ");\n"
 
 --Helper function that escapes characters in strings
