@@ -2,19 +2,6 @@
     BNF Converter: OCaml main file
     Copyright (C) 2005  Author:  Kristofer Johannisson
 
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
 -}
 
 -- based on BNFC Haskell backend
@@ -24,44 +11,53 @@ module BNFC.Backend.OCaml (makeOCaml) where
 
 import System.FilePath (pathSeparator, (</>))
 
-import BNFC.Backend.Base hiding (Backend)
+import BNFC.Backend.Base                    (MkFiles, mkfile)
 import BNFC.Backend.Common.Makefile
 import BNFC.Backend.OCaml.CFtoOCamlAbs
 import BNFC.Backend.OCaml.CFtoOCamlLex
 import BNFC.Backend.OCaml.CFtoOCamlPrinter
 import BNFC.Backend.OCaml.CFtoOCamlShow
 import BNFC.Backend.OCaml.CFtoOCamlTemplate
-import BNFC.Backend.OCaml.CFtoOCamlTest
+import BNFC.Backend.OCaml.CFtoOCamlTest     (ocamlTestfile)
 import BNFC.Backend.OCaml.CFtoOCamlYacc
-import BNFC.Backend.XML
+import BNFC.Backend.OCaml.OCamlUtil
+import BNFC.Backend.XML                     (makeXML)
 import BNFC.CF
 import BNFC.Options
 import BNFC.PrettyPrint
 import BNFC.Utils
 
+import qualified BNFC.Backend.C as C
+
 -- naming conventions
 
 noLang :: SharedOptions -> String -> String
 noLang _ name = name
 
 withLang :: SharedOptions -> String -> String
-withLang opts name = name ++ lang opts
+withLang opts name = name ++ sanitizedLang opts
 
 mkMod :: (SharedOptions -> String -> String) -> String -> SharedOptions -> String
 mkMod addLang name opts =
-    pref ++ if inDir opts then lang opts ++ "." ++ name else addLang opts name
+    pref ++ if inDir opts then sanitizedLang opts ++ "." ++ name else addLang opts name
         where pref = maybe "" (++".") (inPackage opts)
 
 mkFile :: (SharedOptions -> String -> String) -> String -> String -> SharedOptions -> FilePath
 mkFile addLang name ext opts =
     pref ++ if inDir opts
-       then lang opts </> name ++ ext'
+       then sanitizedLang opts </> name ++ ext'
        else addLang opts name ++ if null ext then "" else ext'
-    where pref = maybe "" (\p->pkgToDir p </> "") (inPackage opts)
+    where pref = maybe "" (\ p -> pkgToDir p </> "") (inPackage opts)
           ext' = if null ext then "" else "." ++ ext
 
+-- | Turn language name into a valid ocaml module identifier.
+sanitizedLang :: SharedOptions -> String
+sanitizedLang = camelCase_ . lang
+
+
 absFile, absFileM, ocamllexFile, ocamllexFileM, ocamlyaccFile, ocamlyaccFileM,
   utilFile, templateFile, templateFileM, printerFile, printerFileM,
+  showFile, showFileM,
   tFile :: SharedOptions -> String
 absFile       = mkFile withLang "Abs" "ml"
 absFileM      = mkMod  withLang "Abs"
@@ -86,34 +82,38 @@ makeOCaml opts cf = do
       prMod  = printerFileM opts
       showMod = showFileM opts
   do
-    mkfile (absFile opts) $ cf2Abstract absMod cf
-    mkfile (ocamllexFile opts) $ cf2ocamllex lexMod parMod cf
-    mkfile (ocamlyaccFile opts) $
-                 cf2ocamlyacc parMod absMod lexMod  cf
-    mkfile (templateFile opts) $ cf2Template (templateFileM opts) absMod cf
-    mkfile (printerFile opts)  $ cf2Printer prMod absMod cf
-    mkfile (showFile opts)  $ cf2show showMod absMod cf
-    mkfile (tFile opts) $ ocamlTestfile absMod lexMod parMod prMod showMod cf
-    mkfile (utilFile opts) utilM
+    mkfile (absFile opts)       comment $ cf2Abstract absMod cf
+    mkfile (ocamllexFile opts)  comment $ cf2ocamllex lexMod parMod cf
+    mkfile (ocamlyaccFile opts) C.comment $
+      cf2ocamlyacc (ocamlParser opts) absMod cf
+    mkfile (templateFile opts)  comment $ cf2Template (templateFileM opts) absMod cf
+    mkfile (printerFile opts)   comment $ cf2Printer prMod absMod cf
+    mkfile (showFile opts)      comment $ cf2show showMod absMod cf
+    mkfile (tFile opts)         comment $
+      ocamlTestfile (ocamlParser opts) absMod lexMod parMod prMod showMod cf
+    mkfile (utilFile opts)      comment $ utilM
     mkMakefile opts $ makefile opts
     case xml opts of
       2 -> makeXML opts True cf
       1 -> makeXML opts False cf
       _ -> return ()
 
+comment :: String -> String
+comment x = unwords [ "(*", x, "*)" ]
+
 pkgToDir :: String -> FilePath
 pkgToDir = replace '.' pathSeparator
 
 codeDir :: SharedOptions -> FilePath
 codeDir opts = let pref = maybe "" pkgToDir (inPackage opts)
-                   dir = if inDir opts then lang opts else ""
+                   dir = if inDir opts then sanitizedLang opts else ""
                    sep = if null pref || null dir then "" else [pathSeparator]
                  in pref ++ sep ++ dir
 
 makefile :: SharedOptions -> String -> Doc
 makefile opts basename = vcat
     [ mkVar "OCAMLC" "ocamlc"
-    , mkVar "OCAMLYACC" "ocamlyacc"
+    , mkVar "OCAMLYACC" $ ocamlParserName opts
     , mkVar "OCAMLLEX" "ocamllex"
     , mkVar "OCAMLCFLAGS" ""
     , mkRule "all" []
@@ -146,9 +146,7 @@ makefile opts basename = vcat
 
 utilM :: String
 utilM = unlines
-    ["(* automatically generated by BNFC *)",
-     "",
-     "open Lexing",
+    ["open Lexing",
      "",
      "(* this should really be in the parser, but ocamlyacc won't put it in the .mli *)",
      "exception Parse_error of Lexing.position * Lexing.position"
