@@ -6,14 +6,14 @@ import Test.Hspec
 import BNFC.GetCF
 
 import BNFC.CF
-import qualified AbsBNF as Abs
+import qualified BNFC.Abs as Abs
 
 spec :: Spec
 spec = do
   describe "transItem" $ do
 
     it "translate a non-terminal" $
-        transItem (Abs.NTerminal (Abs.IdCat (Abs.Ident "Foo3")))
+        transItem (Abs.NTerminal (Abs.IdCat (npIdentifier "Foo3")))
             `shouldBe` [Left (CoercCat "Foo" 3)]
 
     it "translate a terminal" $
@@ -28,9 +28,8 @@ spec = do
   describe "checkRule" $ do
 
     it "returns an error if the rule uses an unknown category" $ do
-        let rulep = Rule ("Foo", ("Foo", [])) (Cat "Bar") [Left (Cat "Baz")]
-            rule = Rule "Foo" (Cat "Bar") [Left (Cat "Baz")]
-            cf = CFG [] [] [] [] [] [rule]
+        let rule  = npRule "Foo"              (Cat "Bar")              [Left (Cat "Baz")] Parsable
+            cf = CFG [] mempty [] [] [] [] [rule] mempty
             expected =
                 "no production for Baz, appearing in rule\n    Foo. Bar ::= Baz"
-        checkRule cf rulep `shouldBe` Just expected
+        checkRule cf rule `shouldBe` Just expected
