@@ -1,16 +1,15 @@
 {-# LANGUAGE FlexibleInstances, UndecidableInstances #-}
+
 module BNFC.OptionsSpec where
 
-import Control.Monad (liftM, liftM2)
-import Data.List (intercalate)
-import Data.Maybe (fromJust)
-import System.Console.GetOpt
-import System.FilePath ((<.>), takeBaseName)
+import Control.Monad.Writer (WriterT(..))
+import Data.List            (intercalate)
+import System.FilePath      ((<.>), takeBaseName)
+
 import Test.Hspec
 import Test.QuickCheck
 
-import BNFC.Options -- SUT
-import BNFC.WarningM
+import BNFC.Options
 
 
 -- Expectation that a particular option has a particular value
@@ -23,57 +22,61 @@ spec = do
   describe "parseMode" $ do
 
     it "returns Help on an empty list of arguments" $
-      parseMode [] `shouldBe` Help
+      parseMode_ [] `shouldBe` Help
 
     it "returns Help if given --help" $
-      parseMode ["--help"] `shouldBe` Help
+      parseMode_ ["--help"] `shouldBe` Help
 
     it "returns Version if given --version" $
-      parseMode ["--version"] `shouldBe` Version
+      parseMode_ ["--version"] `shouldBe` Version
 
     it "returns an error if help is given an argument" $
-      isUsageError (parseMode ["--help=2"]) `shouldBe` True
+      isUsageError (parseMode_ ["--help=2"]) `shouldBe` True
 
     it "If no language is specified, it should default to haskell" $
-      parseMode["file.cf"] `shouldSet` (target, TargetHaskell)
+      parseMode_ ["file.cf"] `shouldSet` (target, TargetHaskell)
 
     it "returns an error if the grammar file is missing" $
-      parseMode["--haskell"] `shouldBe` UsageError "Missing grammar file"
+      parseMode_ ["--haskell"] `shouldBe` UsageError "Missing grammar file"
 
     it "returns an error if multiple grammar files are given" $
-      parseMode["--haskell", "file1.cf", "file2.cf"]
+      parseMode_ ["--haskell", "file1.cf", "file2.cf"]
         `shouldBe` UsageError "Too many arguments"
 
     it "sets the language name to the basename of the grammar file" $
-      parseMode["foo.cf"] `shouldSet` (lang, "foo")
+      parseMode_ ["foo.cf"] `shouldSet` (lang, "foo")
 
     it "accept 'old style' options" $ do
-      parseMode["-haskell", "-m", "-glr", "file.cf"]
+      parseMode_ ["-haskell", "-m", "-glr", "file.cf"]
         `shouldSet` (target, TargetHaskell)
-      parseMode["-haskell", "-m", "-glr", "file.cf"]
+      parseMode_ ["-haskell", "-m", "-glr", "file.cf"]
         `shouldSet` (make, Just "Makefile")
-      parseMode["-haskell", "-m", "-glr", "file.cf"]
+      parseMode_ ["-haskell", "-m", "-glr", "file.cf"]
         `shouldSet` (glr, GLR)
 
   it "accept latex as a target language" $
-    parseMode["--latex", "file.cf"] `shouldSet` (target, TargetLatex)
+    parseMode_ ["--latex", "file.cf"] `shouldSet` (target, TargetLatex)
 
   describe "Old option translation" $ do
     it "translate -haskell to --haskell" $
-      translateOldOptions ["-haskell"] `shouldBe` ["--haskell"]
+      translateOldOptions ["-haskell"] `shouldBe`
+        (WriterT $ Right (["--haskell"]
+                         ,["Warning: unrecognized option -haskell treated as if --haskell was provided."]))
 
     describe "--makefile" $ do
 
       it "is off by default" $
-        parseMode["--c", "foo.cf"] `shouldSet` (make, Nothing)
+        parseMode_ ["--c", "foo.cf"] `shouldSet` (make, Nothing)
 
       it "uses the file name 'Makefile' by default" $
-        parseMode["--c", "-m", "foo.cf"] `shouldSet` (make, Just "Makefile")
+        parseMode_ ["--c", "-m", "foo.cf"] `shouldSet` (make, Just "Makefile")
 
       context "when using the option with an argument" $
         it "uses the argument as Makefile name" $
-          parseMode["--c", "-mMyMakefile", "foo.cf"]
+          parseMode_ ["--c", "-mMyMakefile", "foo.cf"]
             `shouldSet` (make, Just "MyMakefile")
+  where
+  parseMode_ = fst . parseMode
 
 isUsageError :: Mode -> Bool
 isUsageError = \case
@@ -85,7 +88,7 @@ isUsageError = \case
 randomOption :: Gen String
 randomOption = oneof [ nonOption, noArg, withArg ]
   where nonOption = stringOf1 ['a'..'z'] -- non-option argument
-        noArg     = liftM ("--"++) nonOption -- flag
+        noArg     = ("--"++) <$> nonOption -- flag
         withArg   = do
           arg   <- nonOption
           flag  <- noArg
@@ -102,7 +105,8 @@ stringOf1 :: [Char] -> Gen String
 stringOf1 = listOf1 . elements
 
 
-instance Arbitrary Target where arbitrary = elements [minBound .. ]
+instance Arbitrary Target where
+  arbitrary = elements [minBound .. maxBound]
 
 -- creates a filepath with the given extension
 arbitraryFilePath :: String -> Gen FilePath
@@ -116,15 +120,16 @@ arbitraryFilePath ext = do
 -- going to generate accidentally an global/target language option
 arbitraryOption :: Gen String
 arbitraryOption = oneof [arbitraryShortOption, arbitraryLongOption]
-  where arbitraryShortOption = liftM (('-':) . (:[])) (elements ['x'..'z'])
-        arbitraryLongOption  = liftM ("--" ++) (stringOf1 ['x'..'z'])
+  where
+  arbitraryShortOption = ('-':) . (:[]) <$> elements ['x'..'z']
+  arbitraryLongOption  = ("--" ++)      <$> stringOf1 ['x'..'z']
 
 -- Arbitrary instance for Mode
 instance Arbitrary Mode where
   arbitrary = oneof
     [ return Help
     , return Version
-    , liftM UsageError arbitrary          -- generates a random error message
+    , UsageError <$> arbitrary          -- generates a random error message
     , do target' <- arbitrary              -- random target
          cfFile <- arbitraryFilePath "cf"
          let args = defaultOptions { lang = takeBaseName cfFile, target = target'}
